#cython: boundscheck=False, wraparound=False, initializedcheck=False

"""The data structures, and related algorithms, which form the basis of the beam tracing
library.

Details on each class, method and function in this sub-module are provided mostly for
developers. Users should refer to :ref:`tracing_manual` for details on beam tracing,
:meth:`.Model.beam_trace` for the main method through which beam traces can be performed
on a model and :mod:`.tracing.tools` for the various beam propagation tools which the
beam tracing library provides.
"""

from finesse.cymath cimport complex_t
from finesse.cymath.complex cimport conj
from finesse.cymath.gaussbeam cimport (
    transform_q,
    inv_transform_q,
    abcd_multiply,
    sym_abcd_multiply,
)
from finesse.cymath.math cimport float_eq

cimport numpy as np
import numpy as np

from copy import copy
from itertools import chain
import logging

import networkx as nx

from finesse.gaussian import BeamParam, transform_beam_param
from finesse.exceptions import TotalReflectionError
from finesse.utilities import refractive_index


LOGGER = logging.getLogger(__name__)


cdef class TraceForest:
    """A container structure which stores multiple :class:`.TraceTree` instances.

    The :class:`.Model` stores a TraceForest object which then represents the current
    tracing state of the configuration. Each time a :meth:`.Model.beam_trace` is called,
    either directly or indirectly, the TraceForest of the Model will be used to perform
    the tracing via propagation of the beam parameters through each tree. This is also
    detailed in :ref:`tracing_manual`.

    Determination of the ordering and overall structure of the TraceForest happens through
    the "planting" of the forest. By calling :meth:`.TraceForest.plant`, the forest is cleared
    and re-planted according to the ordered list of trace dependencies passed to this method.
    This is a step which is performed automatically in :meth:`.Model.beam_trace`, where this
    re-planting process only occurs under the following condition:

     * a connector has been added or removed since the last call,
     * the type of beam tracing has been switched from symmetric to
       asymmetric or vice-verase,
     * or the tracing priority (i.e. ordered list of trace dependencies)
       has changed in any way.

    In the initialisation process of building a simulation, a specialised version of a TraceForest
    is constructed from the model TraceForest using the ``TraceForest.make_changing_forest`` method.
    This inspects the model forest and selects only those trees, and branches of trees, which will
    have changing beam parameters during the simulation; i.e. due to some :class:`.GeometricParameter`
    being scanned. This new, "changing TraceForest" is then the optimised structure via which
    simulation-time beam traces (on changing beam parameter paths) are performed. More details on
    this, including additional powerful features that this changing forest provides, can be found
    in :ref:`tracing_manual`.
    """
    def __init__(self, object model):
        self.forest = []
        self.N_trees = 0

        self.dependencies = []

        self.symmetric = False

        self.model = model

        self.node_coverage = []
        self.internal_cavity_nodes = []

    def __deepcopy__(self, memo):
        raise RuntimeError("TraceForest instances cannot be copied.")

    def plant(self, list trace_order):
        """Constructs and stores all the trace trees according to
        the order of dependencies in `trace_order`.

        Parameters
        ----------
        trace_order : list
            List of the dependency objects by priority of tracing.
        """
        from finesse.components import Cavity

        self.clear()
        self.dependencies = trace_order.copy()

        # Always make the internal cavity trees first, these
        # will be generated in the order in which each cavity
        # appears in the trace_order dependency list
        cavities = list(filter(lambda x: isinstance(x, Cavity), self.dependencies))
        self._add_internal_cavity_trees(cavities)
        LOGGER.debug("TraceForest with internal Cavity trees: %s", self)

        # Handle ordering of internal cavity trees which overlap
        self._handle_overlapping_cavities()
        LOGGER.debug("TraceForest after overlapping cavities handled: %s", self)

        # Iterate through all dependencies by tracing order
        # and generate their associated trees
        for dep in self.dependencies:
            if isinstance(dep, Cavity):
                self._add_external_cavity_tree(dep)
            else:
                self._add_gauss_tree(dep)
        LOGGER.debug(
            "TraceForest with internal, external Cavity trees and Gauss trees: %s",
            self,
        )

        # Remove common sub-trees sequentially
        self.trim()
        LOGGER.debug("TraceForest after trimming: %s", self)
        # Update the node coverage list to take account
        # for changes made during trim
        self.node_coverage = list(self.gather_nodes())
        # Add branch trees at beam splitters where the nodes weren't
        # reachable from a previous trace tree
        cdef int missing = self._add_beamsplitter_branch_trees()
        LOGGER.debug("TraceForest after branch trees planted: %s", self)

        # If we're not symmetric and there are nodes which we can't
        # reach because of this (i.e. no path leading back to these
        # nodes), then we need to rectify this
        if not self.symmetric and missing:
            self._add_backwards_nonsymm_trees()
            LOGGER.debug("TraceForest after first backwards, asymmetric trees added: %s", self)
            # It's possible that there are still some branch nodes from beam splitters
            # that we couldn't reach at this point in asymmetric traces (e.g. a typical PDH
            # setup with a pick-off BS and one cav command), so we need to repeat the above
            # all over again to catch these trees
            if self._add_beamsplitter_branch_trees():
                self._add_backwards_nonsymm_trees()
            LOGGER.debug("TraceForest after last backwards, asymmetric trees added: %s", self)


        cdef set diff = self.find_untraversed_nodes()
        if len(diff):
            raise RuntimeError(
                "Bug encountered! The following optical nodes are missing from the "
                f"model trace forest:\n{diff}"
            )

    cpdef void clear(self):
        """Clears the trace forest, removing all trace trees."""
        self.forest.clear()
        self.N_trees = 0
        self.dependencies.clear()
        self.internal_cavity_nodes.clear()
        self.node_coverage.clear()

    def full_beam_trace(self):
        """Performs a "model-time" beam trace on all trace trees.

        This method is called internally by :meth:`.Model.beam_trace`. One should
        use that method to get a more complete representation of the tracing of
        the beam through a model.

        Returns
        -------
        trace : dict
            Dictionary of `node: (qx, qy)` mappings where `node` is each
            :class:`.OpticalNode` instance and `qx, qy` are the beam
            parameters in the tangential and sagittal planes, respectively,
            at these nodes.
        """
        cdef:
            Py_ssize_t tree_idx
            TraceTree tree

            double lambda0 = self.model.lambda0

            object node
            dict trace = {}

        for tree_idx in range(self.N_trees):
            tree = self.forest[tree_idx]

            if tree.is_source:
                node = tree.node
                qx = tree.dependency.qx
                qy = tree.dependency.qy
                trace[node] = qx, qy
                if self.symmetric:
                    trace[node.opposite] = qx.reverse(), qy.reverse()

            propagate(tree, trace, lambda0, self.symmetric)

        return trace

    cdef int _add_internal_cavity_trees(self, list cavities) except -1:
        cdef:
            Py_ssize_t Ncavs = len(cavities)
            Py_ssize_t cav_idx
            object cav

        for cav_idx in range(Ncavs):
            cav = cavities[cav_idx]

            self.forest.append(TraceTree.from_cavity(cav))
            self.node_coverage.extend(cav.path.nodes_only)
            self.internal_cavity_nodes.extend(cav.path.nodes_only)
            self.N_trees += 1

        return 0

    cdef int _handle_overlapping_cavities(self) except -1:
        cdef:
            Py_ssize_t i, j
            TraceTree tree1, tree2

            list tree1_nodes = []
            list overlaps = []

        # Gather all the overlapping cavity combinations
        for i in range(self.N_trees):
            tree1 = self.forest[i]
            tree1_nodes = []
            _gather_nodes(tree1, tree1_nodes)
            for j in range(i + 1, self.N_trees):
                tree2 = self.forest[j]
                for node in tree1_nodes:
                    if tree2.contains_node(node):
                        overlaps.append((tree1, tree2))
                        break

        # If there are no overlaps then nothing to do here
        if not overlaps:
            return 0

        # Make a flattened list of the overlapping tree combinations
        merged = list(chain.from_iterable(overlaps))
        # and merge the column slices of these combinations to get
        # the correct ordering before the operations below
        merged = merged[::2] + merged[1::2]

        # From this, create a list of these unique trees with the order
        # reversed such that trees which were added first from the cavity
        # trace order will overwrite the branches of the later trees which
        # intersect with them. This then guarantees that the trace_order given
        # to TraceForest.plant will be preserved for overlapping cavities.
        new_inner_order = list(dict.fromkeys(reversed(merged)))

        # Remove the internal cavity trees which are present in this
        # overlapping trees container...
        self.forest = [tree for tree in self.forest if tree not in new_inner_order]
        # ... and then add them back in the order as outlined above
        for tree in new_inner_order:
            self.forest.append(tree)

        return 0

    cdef int _add_external_cavity_tree(self, object cav) except -1:
        cdef:
            dict exit_nodes
            object source, target

            TraceTree new_tree

        exit_nodes = cav.get_exit_nodes()
        for source, target in exit_nodes.items():
            new_tree = TraceTree.from_node(
                target,
                cav,
                self.symmetric,
                pre_node=source,
                exclude=self.node_coverage,
            )
            if new_tree is not None:
                self.forest.append(new_tree)
                self.N_trees += 1

                _gather_nodes(new_tree, self.node_coverage)

        return 0

    cdef int _add_gauss_tree(self, object gauss) except -1:
        cdef:
            object gauss_node = gauss.node

            Py_ssize_t tree_idx
            TraceTree tree, new_tree, found, fp, fpp
            TraceTree rm_tree
            set trees_to_remove = set()

        if gauss_node in self.internal_cavity_nodes:
            LOGGER.error(
                "Gauss object %s is at an internal Cavity node (%s). Ignoring "
                "this Gauss. Remove the corresponding Cavity object to "
                "propagate this Gauss object instead.",
                gauss.name,
                gauss_node.full_name,
            )
            return 0

        # Find branches in trees already planted that contain the gauss node
        # or its opposite direction
        for tree_idx in range(self.N_trees):
            tree = self.forest[tree_idx]

            found = tree.find_tree_at_node(gauss_node, include_opposite=True)
            if found is not None: # found a tree at the gauss node
                fp = found.parent
                if fp is not None: # this tree has a parent
                    # the parent node is an output i.e. a space exists between
                    # parent and the gauss -> so we want to remove parent too
                    if not fp.node.is_input:
                        fpp = fp.parent
                        if fpp is None: # parent has no parent so just remove the whole tree
                            trees_to_remove.add(tree)
                        else: # parent has a parent
                            if fp == fpp.left:
                                fpp.remove_left()
                            else:
                                fpp.remove_right()

                            # if the parent of parent has no remaining connections
                            # just remove the whole tree
                            if fpp.left is None and fpp.right is None and fpp.parent is None:
                                trees_to_remove.add(tree)
                    else: # parent node is an input (no space between parent -> gauss)
                        if found == fp.left:
                            fp.remove_left()
                        else:
                            fp.remove_right()

                        if fp.left is None and fp.right is None and fp.parent is None:
                            trees_to_remove.add(tree)
                else: # found tree has no parent so just remove the whole tree
                    trees_to_remove.add(tree)

        new_tree = TraceTree.from_node(
            gauss_node,
            gauss,
            self.symmetric,
            is_gauss=True,
            exclude=self.internal_cavity_nodes
        )
        if new_tree is not None:
            self.forest.append(new_tree)
            self.N_trees += 1

            _gather_nodes(new_tree, self.node_coverage)

        # if the opposite direction of gauss node is not in the forward propagated
        # gauss tree then we need to make a tree from gauss.opposite too
        if (
            self.symmetric or
            (
                new_tree is not None and
                new_tree.find_tree_at_node(gauss_node.opposite) is None
            )
        ):
            # backwards tree is not from the gauss node itself now
            # so leave is_gauss as False otherwise there would be
            # two sources from the same gauss node leading to problems
            back_new_tree = TraceTree.from_node(
                gauss_node.opposite,
                gauss,
                self.symmetric,
                exclude=self.internal_cavity_nodes
            )

            if (
                back_new_tree is not None and
                (back_new_tree.left is not None or back_new_tree.right is not None)
            ):
                self.forest.append(back_new_tree)
                self.N_trees += 1

                _gather_nodes(back_new_tree, self.node_coverage)

        for rm_tree in trees_to_remove:
            self.forest.remove(rm_tree)
            self.N_trees -= 1

        return 0

    cdef int _add_beamsplitter_branch_trees(self) except -1:
        cdef:
            double node_nr, pre_node_nr
            set diff = self.find_untraversed_nodes()
            dict branch_start_nodes = {}

            TraceTree new_tree

        from finesse.components.general import InteractionType

        if not diff:
            return 0

        for node in diff:
            pre = list(self.model.optical_network.predecessors(node.full_name))
            node_nr = refractive_index(node)
            for pre_node_name in pre:
                pre_node = self.node_from_name(pre_node_name)
                pre_node_nr = refractive_index(pre_node)
                # if predecessor node is also in the set of unreachable nodes OR
                # the interaction type from pre_node -> node is a reflection and
                # the refractive indices at these ports are not the same (total
                # internal reflection) then skip it
                if (
                    pre_node in diff
                    or (
                        node.component.interaction_type(pre_node, node) == InteractionType.REFLECTION
                        and not float_eq(node_nr, pre_node_nr)
                    )
                ):
                    continue

                # otherwise we want to store the predecessor node and its associated
                # dependency in a dict to make trees from later
                branch_start_nodes[node] = pre_node, self.find_dependency_from_node(pre_node)
                break

        if not branch_start_nodes:
            # Shouldn't ever have a case where a branched node does not have
            # a predecessor (when doing symmetric planting), so if this
            # happens a bug has been encountered
            if self.symmetric:
                raise RuntimeError(
                    "Bug encountered! Could not create branch trees from "
                    f"the following missed nodes: {diff}"
                )
            # But if we're not a symmetric forest, then this will occur in
            # most cases as (for anything but extremely simple files) there
            # will be nodes which can't be reached directly, so inform on
            # return that this is the case
            else:
                return 1

        for node, (pre_node, dependency) in branch_start_nodes.items():
            new_tree = TraceTree.from_node(
                node,
                dependency,
                self.symmetric,
                pre_node=pre_node,
                exclude=self.node_coverage,
            )
            if new_tree is not None:
                self.forest.append(new_tree)
                self.N_trees += 1
                _gather_nodes(new_tree, self.node_coverage)

        return self._add_beamsplitter_branch_trees()

    cdef int _add_backwards_nonsymm_trees(self) except -1:
        cdef:
            TraceTree tree
            # twisted tree root and branch
            TraceTree ttr, ttb

            object comp
            bint is_dependency_changing

        # Get all the nodes we can't reach due to the asymmetric trace
        unreachable_nodes = self.find_untraversed_nodes()
        for node in unreachable_nodes:
            tree = self.find_tree_from_node(node.opposite)
            if tree is None:
                continue

            is_dependency_changing = tree.dependency.is_changing
            # Here we begin the process of making a "twisted tree" where
            # the left sub-tree node is actually a pre-coupling of the
            # parent tree node, this will allow us to apply the inverse
            # ABCD law transformation to left sub-tree node during tracing
            ttr = TraceTree.initialise(
                tree.node, tree.dependency, &is_dependency_changing
            )
            ttb = TraceTree.initialise(
                node, tree.dependency, &is_dependency_changing
            )

            if ttb.node.is_input:
                comp = ttb.node.component
            else:
                comp = ttb.node.space

            ttr.left = ttb
            ttb.parent = ttr

            try:
                # Check that there is a coupling from unreachable node -> opposite
                comp.check_coupling(ttb.node, ttr.node)

                try:
                    # Just to re-iterate, here we get the ABCD matrix coupling from
                    # the left tree to the parent tree (opposite to usual) in order
                    # to use this matrix in the inverse ABCD law transformation
                    ttr.sym_left_abcd_x, ttr.left_abcd_x = comp.ABCD(
                        ttb.node, ttr.node, direction="x", copy=False, retboth=True,
                    )
                    ttr.sym_left_abcd_y, ttr.left_abcd_y = comp.ABCD(
                        ttb.node, ttr.node, direction="y", copy=False, retboth=True,
                    )
                except TotalReflectionError:
                    raise

                ttr.is_left_surf_refl = is_surface_refl(comp, ttb.node, ttr.node)

                ttb.is_x_changing |= (
                    ttr.sym_left_abcd_x is not None
                    and is_abcd_changing(ttr.sym_left_abcd_x)
                )
                ttb.is_y_changing |= (
                    ttr.sym_left_abcd_y is not None
                    and is_abcd_changing(ttr.sym_left_abcd_y)
                )

                # Now mark the twisted tree root as an inverse transformation tree
                ttr.do_inv_transform = True
            except ValueError:
                # No coupling exists from ttb.node -> ttr.node (typically means no
                # reflection coupling) so mark the root as needing a -q* operation
                # instead now as there is no other way to set the node q otherwise
                ttr.do_nonsymm_reverse = True

            self.forest.append(ttr)
            self.N_trees += 1
            _gather_nodes(ttr, self.node_coverage)

    cpdef set find_untraversed_nodes(self):
        """Finds all the optical nodes in the model which are not
        covered by the trace forest."""
        cdef:
            set nodes_traversed = set(self.node_coverage)
            set nominal_diff = set(self.model.optical_nodes).difference(nodes_traversed)
            set real_diff = set()

        if not self.symmetric:
            return nominal_diff

        for node in nominal_diff:
            if node.opposite not in nodes_traversed:
                real_diff.add(node)

        return real_diff

    cpdef TraceTree find_tree_from_node(self, object node):
        """Given an optical node, this finds the :class:`.TraceTree` instance
        corresponding to this node (if one exists).

        Parameters
        ----------
        node : :class:`.OpticalNode`
            An optical node.

        Returns
        -------
        tree : :class:`.TraceTree`
            The tree corresponding to `node`, or ``None`` if none found.
        """
        cdef:
            Py_ssize_t tree_idx
            TraceTree tree, found

        for tree_idx in range(self.N_trees):
            tree = self.forest[tree_idx]

            found = tree.find_tree_at_node(node, self.symmetric)
            if found is not None:
                return found

        return None

    cpdef object find_dependency_from_node(self, object node, bint raise_not_found=True):
        """Finds the dependency object associated with the optical `node`.

        If no tree is found associated with this node, and `raise_not_found` is true,
        then a ``RuntimeError`` is raised. Otherwise `None` is returned.

        Parameters
        ----------
        node : :class:`.OpticalNode`
            An optical node.

        raise_not_found : bool, optional; default: True
            Raises a RuntimeError if no dependency found. Returns `None` if False.
        """
        cdef:
            Py_ssize_t tree_idx
            TraceTree tree, found

        for tree_idx in range(self.N_trees):
            tree = self.forest[tree_idx]

            found = tree.find_tree_at_node(node, self.symmetric)
            if found is not None:
                return found.dependency

        if raise_not_found:
            raise RuntimeError(
                "Bug encountered! Could not find a dependency object "
                f"associated with node {node.full_name} in the trace forest."
            )
        else:
            return None

    cpdef set gather_nodes(self):
        """Generates a set of all the optical nodes covered by the forest."""
        cdef:
            Py_ssize_t tree_idx
            TraceTree tree

            list gather = []

        for tree_idx in range(self.N_trees):
            tree = self.forest[tree_idx]

            _gather_nodes(tree, gather)

        return set(gather)

    cdef object node_from_name(self, name):
        return self.model.network.nodes[name]["weakref"]()

    cdef TraceForest make_changing_forest(self):
        """Constructs a new TraceForest from this forest, consisting
        of only the trees which will have changing beam parameters.

        This method is called in BaseSimulation._initialise for setting
        up the simulation trace forest used for efficient beam tracing.
        """
        cdef:
            Py_ssize_t tree_idx
            TraceTree tree, chtree

            list changing_trees = []
            bint branch_added = False
            TraceForest changing_forest = TraceForest(self.model)

        for tree_idx in range(self.N_trees):
            tree = self.forest[tree_idx]
            branch_added = False
            # For branched beamsplitter trees, need to see if the root
            # node opposite is already present in the changing trees
            # list --- if so then add the branch tree as this will also
            # be changing
            for chtree in changing_trees:
                if chtree.find_tree_at_node(tree.node.opposite) is not None:
                    changing_trees.append(tree)
                    branch_added = True
                    break

            if not branch_added:
                # From this tree, obtain the broadest sub-trees which
                # will have changing beam parameters
                changing_trees.extend(tree.get_broadest_changing_subtrees())

        cdef Py_ssize_t Nchanging_nominal = len(changing_trees)
        cdef set roots = set() # Parents of changing trees
        cdef set trees_to_remove = set()
        for tree_idx in range(Nchanging_nominal):
            tree = changing_trees[tree_idx]
            # Now set each changing tree to the parent tree so that
            # the root is used in beam tracing...
            if tree.parent is not None:
                # ... but only do this for parents not yet added
                # otherwise could get duplicate trees
                if tree.parent not in roots:
                    changing_trees[tree_idx] = tree.parent
                    roots.add(tree.parent)
                else:
                    # This tree will already be encapsulated by the previous
                    # parent so just mark it to be removed
                    trees_to_remove.add(tree)

        for tree in trees_to_remove:
            changing_trees.remove(tree)

        changing_forest.forest = changing_trees
        changing_forest.N_trees = len(changing_trees)
        changing_forest.symmetric = self.symmetric
        changing_forest.internal_cavity_nodes = self.internal_cavity_nodes

        return changing_forest

    cpdef bint contains_space(self, object space):
        """Whether this trace forest contains the specified space.

        Parameters
        ----------
        space : :class:`.Space`
            A space instance.

        Returns
        -------
        flag : bool
            True if the forest contains `space`, False otherwise.
        """
        cdef:
            Py_ssize_t tree_idx
            TraceTree tree

        for tree_idx in range(self.N_trees):
            tree = self.forest[tree_idx]

            if tree.contains_space(space):
                return True

        return False

    cpdef bint contains_comp(self, object comp):
        cdef:
            Py_ssize_t tree_idx
            TraceTree tree

        for tree_idx in range(self.N_trees):
            tree = self.forest[tree_idx]

            if tree.contains_comp(comp):
                return True

        return False

    cpdef tuple find_potential_mismatch_couplings(self, TraceForest other=None):
        """Retrieves the node couplings which are potentially mode mismatched. If
        `other` is not given then the couplings which are local to this forest only
        will be found, otherwise couplings between this forest and `other` will
        be retrieved.

        If this forest is asymmetric, then calling this method is equivalent to
        calling :meth:`.TraceForest.find_intersection_couplings`.

        This method is used internally for obtaining all the possible mode mismatch
        couplings between a changing trace forest (held by a modal simulation) and
        the main model trace forest.

        Parameters
        ----------
        other : :class:`.TraceForest`
            Find dependencies from a different trace forest than this one
            when checking for mode mismatch couplings.

        Returns
        -------
        couplings : tuple
            A tuple of the node couplings where each element is ``(from_node, to_node)``.
        """
        from finesse.components.general import InteractionType

        intersect_couplings = self.find_intersection_couplings(other)

        cdef list refls = []
        cdef list fake_refl_mismatches = []
        cdef tuple other_mrefls
        cdef object[:, ::1] refl_abcd_sym_x
        cdef object[:, ::1] refl_abcd_sym_y
        # If we're doing a symmetric trace then self-reflections from mirror-type components
        # can be potential mode mismatch couplings so need to add these too
        if self.symmetric:
            refls.extend(self.get_mirror_reflection_couplings())
            for node1, node2 in refls:
                opp_surface_onode = None
                # Find the output node on the other side of the surface
                for n1s_name in list(self.model.optical_network.successors(node1.full_name)):
                    n1s = self.node_from_name(n1s_name)
                    if node1.component.interaction_type(node1, n1s) == InteractionType.TRANSMISSION:
                        opp_surface_onode = n1s
                        break

                # If none found then nothing else needs to be done for this coupling
                if opp_surface_onode is None:
                    continue

                # Reflection coupling on other side of surface is in the potential
                # mismatch couplings so this one can remain too
                if (opp_surface_onode.opposite, opp_surface_onode) in refls:
                    continue

                # Get the dependencies associated with the two sides of the surface
                dep1 = self.find_dependency_from_node(node1)
                dep2 = self.find_dependency_from_node(opp_surface_onode, raise_not_found=False)
                if dep2 is None and other is not None:
                    dep2 = other.find_dependency_from_node(opp_surface_onode)

                # Finally, if the dependencies of the trees on both sides are the same
                # then this isn't really a mismatch coupling (as beam params on both sides
                # are guaranteed to be mode matched in such a case) so mark it to be removed
                if dep1 is dep2:
                    fake_refl_mismatches.append((node1, node2))

            for fnodes in fake_refl_mismatches:
                refls.remove(fnodes)

            # In addition, if the model trace forest is specified via other
            # then we need to check for reflection couplings here which
            # impinge against connectors with changing ABCDs as these will
            # also be potential mode mismatch couplings
            if other is not None:
                other_mrefls = other.get_mirror_reflection_couplings(
                    skip_dependencies=self.dependencies,
                )

                for node1, node2 in other_mrefls:
                    # Don't add the coupling if we already determined that it
                    # was a "fake" mismatch coupling (see above)
                    if (node1, node2) in fake_refl_mismatches:
                        continue

                    comp = node1.component
                    # Get the symbolic ABCDs upon reflection from the surface...
                    refl_abcd_sym_x = comp.ABCD(
                        node1, node2, "x", copy=False, symbolic=True
                    )
                    refl_abcd_sym_y = comp.ABCD(
                        node1, node2, "y", copy=False, symbolic=True
                    )

                    # ... and check if they're changing, if so we have another
                    # possible mode mismatch coupling which needs to be added
                    if is_abcd_changing(refl_abcd_sym_x) or is_abcd_changing(refl_abcd_sym_y):
                        refls.append((node1, node2))

        return intersect_couplings + tuple(set(refls))


    cpdef tuple find_intersection_couplings(self, TraceForest other=None):
        """Finds the node couplings at which trees with differing trace dependencies intersect.

        Parameters
        ----------
        other : :class:`.TraceForest`
            Find dependencies from a different trace forest than this one
            when checking for intersections.

        Returns
        -------
        couplings : tuple
            A tuple of the node couplings where each element is ``(from_node, to_node)``.
        """
        cdef:
            Py_ssize_t otree_idx, itree_idx
            TraceTree tree

            list couplings = []

        if not self.N_trees:
            return ()

        for otree_idx in range(self.N_trees):
            tree = self.forest[otree_idx]

            # Get the final *input* nodes at the end of the tree, across all branches
            last_nodes = tree.get_last_input_nodes()
            # Need to also add the reverse node of the tree if it's a source
            # tree so that intersections are checked in the other propagation
            if tree.is_source and tree.node.opposite not in last_nodes:
                last_nodes.append(tree.node.opposite)

            for node in last_nodes:
                # Obtain the successor nodes of this (if any) from the network
                succ_nodes = list(self.model.optical_network.successors(node.full_name))

                for snode_name in succ_nodes:
                    snode = self.node_from_name(snode_name)
                    # Then for each successor node find its TraceTree in the forest
                    # and obtain the dependency that this relies upon
                    if other is None:
                        snode_dep = self.find_dependency_from_node(snode)
                    else:
                        snode_dep = other.find_dependency_from_node(snode)

                    # If this dependency is not the same object as the original tree's
                    # dependency then we've found an intersection so add this coupling
                    if snode_dep is not tree.dependency:
                        # Here we do some sanity checks on the coupling we've found
                        comp = node.component
                        scomp = snode.component
                        if node.component is not snode.component:
                            raise RuntimeError(
                                "Bug encountered! Found an intersection coupling "
                                f"{node.full_name} -> {snode.full_name} which "
                                "does not occur across the same connector."
                            )
                        node.component.check_coupling(node, snode)

                        couplings.append((node, snode))
                        # Add the reverse coupling too if it exists
                        try:
                            node.component.check_coupling(snode.opposite, node.opposite)
                            couplings.append((snode.opposite, node.opposite))
                        except:
                            pass

        return tuple(set(couplings))

    cpdef tuple get_mirror_reflection_couplings(
        self,
        bint ignore_internal_cavities=True,
        list skip_dependencies=None,
    ):
        """Get the node couplings in the forest which correspond to self-reflections
        from mirror-like components.

        Parameters
        ----------
        ignore_internal_cavities : bool, default: True
            Ignore the node couplings inside cavities.

        skip_dependencies : list
            Optional list of trees to skip based on their dependencies.

        Returns
        -------
        couplings : tuple
            A sequence of tuples consisting of the node1 -> node2 self
            reflection couplings.
        """
        cdef:
            Py_ssize_t tree_idx
            TraceTree tree

            list couplings = []

        from finesse.components import Cavity

        if skip_dependencies is None:
            skip_dependencies = []

        for tree_idx in range(self.N_trees):
            tree = self.forest[tree_idx]

            if tree.dependency in skip_dependencies:
                continue

            if ignore_internal_cavities:
                # The tree is an internal cavity tree itself so skip the whole thing
                if tree.is_source and isinstance(tree.dependency, Cavity):
                    continue

                # If it's a tree coming directly from an internal cavity tree then
                # get the reflection couplings only from the second node onwards
                if tree.node in self.internal_cavity_nodes:
                    if tree.left is not None:
                        couplings.extend(tree.left.get_mirror_reflection_couplings())
                else:
                    couplings.extend(tree.get_mirror_reflection_couplings())
            else:
                couplings.extend(tree.get_mirror_reflection_couplings())

        return tuple(set(couplings))

    cdef void trim(self):
        cdef:
            Py_ssize_t i, j
            TraceTree tree1, tree2

            list tree1_nodes = []
            set trees_to_remove = set()

        from finesse.components import Gauss

        for i in range(self.N_trees):
            tree1 = self.forest[i]
            tree1_nodes = []
            _gather_nodes(tree1, tree1_nodes)

            for j in range(i + 1, self.N_trees):
                tree2 = self.forest[j]
                # Don't attempt to trim common nodes for internal cavity trees
                if not tree2.is_source or isinstance(tree2.dependency, Gauss):
                    # If any parts of tree2 overlap with tree1 then trim off
                    # these branches of tree2 to ensure uniqueness of trees
                    tree2.trim_at_nodes(tree1_nodes, self.symmetric)

                    if not tree2.is_source:
                        # Tree has no branches left so can be removed entirely
                        if tree2.left is None and tree2.right is None:
                            trees_to_remove.add(tree2)

        for tree in trees_to_remove:
            self.forest.remove(tree)
            self.N_trees -= 1

    cpdef draw_by_dependency(self):
        """Draws the forest as a string representation.

        All the trees in the forest are sorted by their dependency and
        stored in the resultant string by these dependency sub-headings. Each
        tree also has its index (i.e. tracing priority) stored in the string
        above the drawn tree.

        Returns
        -------
        forest_str : str
            A string representation of the forest, sorted by dependency with
            tracing priority indices displayed for each tree.
        """
        cdef:
            Py_ssize_t tree_idx
            TraceTree tree

            list dependencies = []

            all_trees_str = ""

        for tree_idx in range(self.N_trees):
            tree = self.forest[tree_idx]
            if tree.dependency not in dependencies:
                dependencies.append(tree.dependency)

        for dependency in dependencies:
            # Make sub-heading for each dependency, giving its name and type name
            all_trees_str += f"\nDependency: {dependency.name} [{type(dependency).__name__}]\n\n"
            for tree_idx in range(self.N_trees):
                tree = self.forest[tree_idx]
                if tree.dependency == dependency:
                    # Give the index of the tree in the forest before drawing each tree
                    # -> indicates tracing order of the tree
                    all_trees_str += (
                        f"  (Index: {tree_idx})\n" + tree.draw(left_pad="    ") + "\n\n"
                    )

        return all_trees_str

    cpdef draw(self):
        """Draws the forest, by trace priority, as a string representation.

        The order in which trees appear in this string represents the order in which
        they will be traced during the beam tracing algorithm.

        In the rare cases where a subsequent tree contains a duplicate node (from an
        earlier tree), the latter tree trace will overwrite the former. This is only
        applicable to configurations with overlapping cavities, and this overwriting
        behaviour will take account of the desired cavity ordering given by the user.

        Returns
        -------
        forest_str : str
            A string representation of the ordered forest.
        """
        cdef:
            Py_ssize_t tree_idx
            TraceTree tree

            object last_dep = None
            all_trees_str = ""

        for tree_idx in range(self.N_trees):
            tree = self.forest[tree_idx]
            dep = tree.dependency

            if dep is not last_dep:
                all_trees_str += f"\nDependency: {dep.name} [{type(dep).__name__}]\n\n"
            else:
                all_trees_str += "\n"

            all_trees_str += tree.draw(left_pad="  ") + "\n"

            last_dep = dep

        return all_trees_str

    def __str__(self):
        return self.draw()


cdef class TraceTree:
    """A binary tree data structure representing all the beam tracing
    paths from some root optical node of a model.

    Each instance of this class has a `left` and `right` sub-tree (of the
    class type) and a parent tree. These linked tree attributes can be None. If
    the tree has a left / right sub-tree then the memoryviews `left_abcd_x`,
    `left_abcd_y` etc. will be initialised from the numerical ABCD matrix from
    the tree's optical node to the next tree's optical node.

    Every sub-tree has a `dependency` attribute which is the object that the
    trace tree depends on - either a :class:`.Cavity` or a :class:`.Gauss`
    instance."""
    def __init__(self, object node, object dependency):
        self.parent = None
        self.left = None
        self.right = None

        self.dependency = dependency # the source object (a cavity or gauss)
        self.node = node # the optical node

        self.is_source = False

        # These will get set in TraceTree.initialise afterwards anyway
        self.is_x_changing = False
        self.is_y_changing = False

        self.left_abcd_x = None
        self.left_abcd_y = None
        self.right_abcd_x = None
        self.right_abcd_y = None
        self.is_left_surf_refl = False

        self.sym_left_abcd_x = None
        self.sym_left_abcd_y = None
        self.sym_right_abcd_x = None
        self.sym_right_abcd_y = None

        self.nr = refractive_index(self.node)

        self.node_id = 0
        self.opp_node_id = 0

        self.do_inv_transform = False
        self.do_nonsymm_reverse = False

    def __deepcopy__(self, memo):
        raise RuntimeError("TraceTree instances cannot be copied.")

    @staticmethod
    cdef TraceTree initialise(
        object node, object dependency, bint* is_dependency_changing=NULL
    ):
        cdef TraceTree tree = TraceTree(node, dependency)

        cdef bint is_dep_changing
        if is_dependency_changing == NULL:
            if dependency is None:
                is_dep_changing = False
            else:
                is_dep_changing = dependency.is_changing
        else:
            is_dep_changing = is_dependency_changing[0]

        tree.is_x_changing = is_dep_changing
        tree.is_y_changing = is_dep_changing

        return tree

    @classmethod
    def from_cavity(cls, cavity):
        """Construct a TraceTree from a cavity instance.

        The resulting tree decays to a linked list as it
        just includes the internal path of the cavity.

        Parameters
        ----------
        cavity : :class:`.Cavity`
            The cavity object.

        Returns
        -------
        tree : :class:`.TraceTree`
            The tree representing the internal cavity path.
        """
        cdef:
            list path = cavity.path.nodes_only
            Py_ssize_t num_nodes = len(path)
            TraceTree parent = None
            TraceTree root

            object node

            bint cav_is_changing = cavity.is_changing

        root = TraceTree.initialise(path[0], cavity, &cav_is_changing)
        root.is_source = True
        parent = root

        cdef Py_ssize_t i
        for i in range(1, num_nodes):
            node = path[i]

            parent = parent.add_left(TraceTree.initialise(node, cavity, &cav_is_changing))

        # Add the final reflection ABCDs for last tree node back to source
        # so that round-trip ABCD can be efficiently computed from root
        parent.left_abcd_x = parent.node.component.ABCD(
            parent.node, root.node, direction='x', copy=False
        )
        parent.left_abcd_y = parent.node.component.ABCD(
            parent.node, root.node, direction='y', copy=False
        )

        # co-ordinate system transformation on reflection due
        # to rotation around vertical axis (inversion)
        parent.is_left_surf_refl = is_surface_refl(
            parent.node.component, parent.node, root.node
        )

        return root

    @staticmethod
    def from_path(list path):
        """Construct a TraceTree from a list of optical nodes.

        The resulting tree decays to a linked list as the path
        is 1D - no branches will occur.

        Parameters
        ----------
        path : list
            A list of optical nodes representing the node path. This
            can be obtained from a :class:`.OpticalPath` instance
            by invoking :attr:`.OpticalPath.nodes_only`.

        Returns
        -------
        tree : :class:`.TraceTree`
            The tree representing the node path.
        """
        cdef:
            Py_ssize_t num_nodes = len(path)
            TraceTree parent = None
            TraceTree root

            object node

        if not num_nodes:
            return None

        root = TraceTree.initialise(path[0], None)
        parent = root

        cdef Py_ssize_t i
        for i in range(1, num_nodes):
            node = path[i]

            parent = parent.add_left(TraceTree.initialise(node, None))

        return root

    @classmethod
    def from_node(
        cls,
        node,
        object dependency,
        bint symmetric,
        pre_node=None,
        bint is_gauss=False,
        list exclude=None,
    ):
        """Construct a TraceTree from an optical node root.

        The resulting tree includes all optical node paths traced
        forward from `node`.

        Parameters
        ----------
        node : :class:`.OpticalNode`
            The root node.

        dependency : :class:`.Cavity` or :class:`.Gauss`
            The dependency object - i.e. what the trace sub-trees depend on.

        symmetric : bool
            Flag indicating whether the tree should be constructed assuming
            that opposite node beam parameters will be set via the reverse
            of the original node beam parameter (true indicates this will be
            the case). In practice, this means that the resultant tree will
            not include any duplicate ports.

        pre_node : :class:`.OpticalNode`, optional; default: None
            An optional node to add before the root for the root sub-tree.

        is_gauss : bool, optional; default: False
            Whether the root node is a gauss node or not.

        Returns
        -------
        tree : :class:`.TraceTree`
            The tree of all paths from `node`.
        """
        cdef:
            object model = node._model

            TraceTree new_root = None
            TraceTree root = None
            TraceTree parent

            dict sub_trees = {}

            unicode n_name
            unicode nbr_name
            dict nbrsdict

            set excl = set()

            bint is_dependency_changing = dependency.is_changing

        if pre_node is not None:
            new_root = TraceTree.initialise(pre_node, dependency, &is_dependency_changing)

        if exclude is not None:
            excl = set(exclude)

        node_from_name = lambda n: model.network.nodes[n]["weakref"]()
        tree = nx.bfs_tree(model.optical_network, node.full_name)
        sub_trees = {}
        cdef Py_ssize_t i, j

        for i, (n_name, nbrsdict) in enumerate(tree.adjacency()):
            n = node_from_name(n_name)

            if n in excl or (symmetric and n.opposite in excl):
                if not i:
                    break

                continue

            excl.add(n)

            if not i:
                root = parent = TraceTree.initialise(n, dependency, &is_dependency_changing)
            else:
                parent = sub_trees.get(n_name, None)
                if parent is None:
                    continue

            for j, nbr_name in enumerate(nbrsdict.keys()):
                nbr = node_from_name(nbr_name)

                if nbr in excl or (symmetric and nbr.opposite in excl):
                    continue

                if not j:
                    sub_trees[nbr_name] = parent.add_left(
                        TraceTree.initialise(nbr, dependency, &is_dependency_changing)
                    )
                else:
                    sub_trees[nbr_name] = parent.add_right(
                        TraceTree.initialise(nbr, dependency, &is_dependency_changing)
                    )

        if root is None:
            return None

        if new_root is not None:
            new_root.add_left(root)
            new_root.is_source = is_gauss
        else:
            root.is_source = is_gauss

        return new_root or root

    cpdef TraceTree add_left(self, TraceTree sub_tree):
        """Add a left sub-tree to the tree.

        Parameters
        ----------
        sub_tree : :class:`.TraceTree`
            The tree to add to the left.

        Returns
        -------
        sub_tree : :class:`.TraceTree`
            The same tree that was added. This is useful for
            looping over a single branch of the tree as a parent
            tree can be set to the return of this method on each iteration.
        """
        cdef:
            object comp
            object parent_node = self.node

        self.left = sub_tree
        sub_tree.parent = self

        if parent_node.is_input:
            comp = parent_node.component
        else:
            comp = parent_node.space

        try:
            self.sym_left_abcd_x, self.left_abcd_x = comp.ABCD(
                parent_node, sub_tree.node, direction='x', copy=False, retboth=True
            )
            self.sym_left_abcd_y, self.left_abcd_y = comp.ABCD(
                parent_node, sub_tree.node, direction='y', copy=False, retboth=True
            )
        except TotalReflectionError:
            raise

        # co-ordinate system transformation on reflection due
        # to rotation around vertical axis (inversion)
        self.is_left_surf_refl = is_surface_refl(comp, parent_node, sub_tree.node)

        sub_tree.is_x_changing |= (
            self.sym_left_abcd_x is not None
            and is_abcd_changing(self.sym_left_abcd_x)
        )
        sub_tree.is_y_changing |= (
            self.sym_left_abcd_x is not None
            and is_abcd_changing(self.sym_left_abcd_y)
        )

        return sub_tree

    cpdef TraceTree add_right(self, TraceTree sub_tree):
        """Add a right sub-tree to the tree.

        Parameters
        ----------
        sub_tree : :class:`.TraceTree`
            The tree to add to the right.

        Returns
        -------
        sub_tree : :class:`.TraceTree`
            The same tree that was added. This is useful for
            looping over a single branch of the tree as a parent
            tree can be set to the return of this method on each iteration.
        """
        cdef:
            object comp
            object parent_node = self.node

        self.right = sub_tree
        sub_tree.parent = self

        if parent_node.is_input:
            comp = parent_node.component
        else:
            comp = parent_node.space

        try:
            self.sym_right_abcd_x, self.right_abcd_x = comp.ABCD(
                parent_node, sub_tree.node, direction='x', copy=False, retboth=True
            )
            self.sym_right_abcd_y, self.right_abcd_y = comp.ABCD(
                parent_node, sub_tree.node, direction='y', copy=False, retboth=True
            )
        except TotalReflectionError:
            raise

        sub_tree.is_x_changing |= (
            self.sym_right_abcd_x is not None
            and is_abcd_changing(self.sym_right_abcd_x)
        )
        sub_tree.is_y_changing |= (
            self.sym_right_abcd_y is not None
            and is_abcd_changing(self.sym_right_abcd_y)
        )

        return sub_tree

    cpdef bint remove_left(self):
        if self.left is None:
            return False

        self.left = None
        self.left_abcd_x = None
        self.left_abcd_y = None
        self.is_left_surf_refl = False

        return True

    cpdef bint remove_right(self):
        if self.right is None:
            return False

        self.right = None
        self.right_abcd_x = None
        self.right_abcd_y = None

        return True

    cdef void trim_at_nodes(
        self, list nodes, bint include_opposite=False
    ):
        if self.left is not None:
            if self.left.node in nodes or (include_opposite and self.left.node.opposite in nodes):
                self.remove_left()
            else:
                self.left.trim_at_nodes(nodes, include_opposite)

        if self.right is not None:
            if self.right.node in nodes or (include_opposite and self.right.node.opposite in nodes):
                self.remove_right()
            else:
                self.right.trim_at_nodes(nodes, include_opposite)

    cdef TraceTree _propagate_find_tree_at_node(
        self,
        TraceTree tree,
        object node,
        bint include_opposite
    ):
        if tree.node == node or (include_opposite and tree.node.opposite == node):
            return tree

        if tree.left is not None:
            lftree = self._propagate_find_tree_at_node(tree.left, node, include_opposite)
            if lftree is not None:
                return lftree

        if tree.right is not None:
            rftree = self._propagate_find_tree_at_node(tree.right, node, include_opposite)
            if rftree is not None:
                return rftree

        return None

    cpdef TraceTree find_tree_at_node(self, object node, bint include_opposite=False):
        return self._propagate_find_tree_at_node(self, node, include_opposite)

    cdef _get_last_input_nodes(self, list last_nodes):
        if self.left is None and self.right is None:
            if self.node.is_input:
                last_nodes.append(self.node)
            else:
                last_nodes.append(self.parent.node)
                # If we're at a beamsplitter then get opposite node
                # of reflection to last node so that both transmission
                # intersections can be picked up properly
                # TODO (sjr) Not sure if this is appropriate for asymmetric
                #            tracing yet, need to think about it
                if self.node.port is not self.parent.node.port:
                    last_nodes.append(self.node.opposite)

        if self.left is not None:
            self.left._get_last_input_nodes(last_nodes)

        if self.right is not None:
            self.right._get_last_input_nodes(last_nodes)

    cpdef list get_last_input_nodes(self):
        cdef:
            list last_nodes = []

        if self.left is None and self.right is None:
            last_nodes.append(self.node)

        if self.left is not None:
            self.left._get_last_input_nodes(last_nodes)

        if self.right is not None:
            self.right._get_last_input_nodes(last_nodes)

        return last_nodes

    cpdef bint contains_node(self, object node):
        return self.find_tree_at_node(node) is not None

    cpdef bint contains_space(self, object space):
        if not self.node.is_input:
            tspace = self.node.space
            if tspace is space:
                return True

        cdef bint left_contains_space = False
        if self.left is not None:
            left_contains_space = self.left.contains_space(space)

        cdef bint right_contains_space = False
        if self.right is not None:
            right_contains_space = self.right.contains_space(space)

        return left_contains_space or right_contains_space

    cpdef bint contains_comp(self, object comp):
        for node in comp.optical_nodes:
            if self.contains_node(node):
                return True

        return False

    cdef __append_mirror_refl_coupling(self, list couplings):
        from finesse.components.surface import Surface

        opp_node = self.node.opposite
        comp = self.node.component
        if opp_node.port is self.node.port and isinstance(comp, Surface):
            # Make sure both sides of the surface get included if encountered
            if self.node.is_input:
                node1 = self.node
                node2 = opp_node
            else:
                node1 = opp_node
                node2 = self.node

            # Check that the reflection coupling makes sense
            if comp.is_valid_coupling(node1, node2):
                couplings.append((node1, node2))

    cdef _get_mirror_refl_couplings(self, list couplings):
        self.__append_mirror_refl_coupling(couplings)

        if self.left is not None:
            self.left._get_mirror_refl_couplings(couplings)

        if self.right is not None:
            self.right._get_mirror_refl_couplings(couplings)

    cpdef list get_mirror_reflection_couplings(self):
        cdef list couplings = []

        self.__append_mirror_refl_coupling(couplings)

        if self.left is not None:
            self.left._get_mirror_refl_couplings(couplings)

        if self.right is not None:
            self.right._get_mirror_refl_couplings(couplings)

        return couplings

    @staticmethod
    cdef bint subtrees_changing(TraceTree sub_tree):
        cdef TraceTree child

        for child in (sub_tree.left, sub_tree.right):
            if child is None:
                continue

            if child.is_x_changing or child.is_y_changing:
                return True

            if TraceTree.subtrees_changing(child):
                return True

        return False

    cpdef bint is_changing(self, bint recursive=True):
        if not recursive:
            return self.is_x_changing or self.is_y_changing

        if self.is_x_changing or self.is_y_changing:
            return True

        return TraceTree.subtrees_changing(self)

    cdef list get_broadest_changing_subtrees(self):
        cdef:
            list changing_trees = []

        _gather_changing_sub_trees(self, changing_trees)
        return changing_trees

    cpdef draw(self, unicode left_pad=""):
        cdef:
            unicode first = self.node.full_name
            list lines = [left_pad + "o" + " " + first]

        _draw_tree(self, "", lines)
        treestr = f"\n{left_pad}".join(lines)

        return treestr

    def __str__(self):
        return self.draw()

    cpdef trace_beam(self, dict trace, double lambda0, bint symmetric):
        """Trace the beam through the tree, placing the beam parameter
        entries at each node in the `trace` dict.
        """
        cdef:
            object node

        if self.is_source:
            node = self.node
            qx = self.dependency.qx
            qy = self.dependency.qy
            trace[node] = qx, qy
            if symmetric:
                trace[node.opposite] = qx.reverse(), qy.reverse()

        propagate(self, trace, lambda0, symmetric)


cdef void _draw_tree(TraceTree tree, unicode lpad, list lines):
    cdef:
        unicode pad
        unicode branch = "├─"
        unicode pipe = "│"
        unicode end = "╰─"
        unicode dash = "─"

        TraceTree ltree = tree.left
        TraceTree rtree = tree.right

    if ltree is not None:
        s = branch + dash
        if rtree is None:
            s = end + dash
            pad = "   "
        else:
            s = branch + dash
            pad = pipe + "   "
        lines.append(lpad + s + "o" + " " + ltree.node.full_name)
        _draw_tree(ltree, lpad + pad, lines)

    if rtree is not None:
        s = end + dash
        pad = "   "
        lines.append(lpad + s + "o" + " " + rtree.node.full_name)
        _draw_tree(rtree, lpad + pad, lines)

cdef void _gather_nodes(TraceTree tree, list gather):
    gather.append(tree.node)

    if tree.left is not None:
        _gather_nodes(tree.left, gather)

    if tree.right is not None:
        _gather_nodes(tree.right, gather)

cdef void _gather_changing_sub_trees(TraceTree tree, list changing_trees):
    cdef:
        TraceTree ltree = tree.left
        TraceTree rtree = tree.right

    if tree.is_changing(recursive=False):
        changing_trees.append(tree)
        return

    if ltree is not None:
        _gather_changing_sub_trees(ltree, changing_trees)
    if rtree is not None:
        _gather_changing_sub_trees(rtree, changing_trees)


cpdef TraceTree get_last_left_branch(TraceTree tree):
    while tree.left is not None:
        if tree.left is None:
            break

        tree = tree.left

    return tree


cpdef bint is_abcd_changing(object[:, ::1] M):
    cdef:
        Py_ssize_t i, j
        bint is_changing

    for i in range(2):
        for j in range(2):
            is_changing = getattr(M[i][j], "is_changing", False)
            if is_changing:
                return True

    return False


cpdef void update_rt_abcd(
    TraceTree tree,
    double[:, ::1] ABCDx,
    double[:, ::1] ABCDy
):
    cdef:
        TraceTree t = tree
        Py_ssize_t i, j

    # Reset round trip ABCD to identity
    for i in range(2):
        for j in range(2):
            if i == j:
                ABCDx[i][j] = 1.0
                ABCDy[i][j] = 1.0
            else:
                ABCDx[i][j] = 0.0
                ABCDy[i][j] = 0.0

    # Find the bottom first as round-trip matrix is
    # computed from multiplying each ABCD "upwards"
    # in the internal tree
    while t.left is not None:
        if t.left is None:
            break

        t = t.left

    while t is not None:
        abcd_multiply(ABCDx, t.left_abcd_x, out=ABCDx, m2_refl_transform=t.is_left_surf_refl)
        abcd_multiply(ABCDy, t.left_abcd_y, out=ABCDy)

        t = t.parent


def update_single_rt_abcd(
    TraceTree tree,
    double[:, ::1] ABCD,
    unicode direction
):
    cdef:
        TraceTree t = tree
        Py_ssize_t i, j
        bint is_x_plane = direction == 'x'

    # Reset round trip ABCD to identity
    for i in range(2):
        for j in range(2):
            if i == j:
                ABCD[i][j] = 1.0
            else:
                ABCD[i][j] = 0.0

    while t.left is not None:
        if t.left is None:
            break

        t = t.left

    while t is not None:
        if is_x_plane:
            abcd_multiply(ABCD, t.left_abcd_x, out=ABCD, m2_refl_transform=t.is_left_surf_refl)
        else:
            abcd_multiply(ABCD, t.left_abcd_y, out=ABCD)

        t = t.parent


cdef void propagate(TraceTree tree, dict trace, double lambda0, bint symmetric):
    cdef:
        object this_node = tree.node
        TraceTree ltree = tree.left
        TraceTree rtree = tree.right
        object left_node = None if ltree is None else ltree.node
        object right_node = None if rtree is None else rtree.node

        tuple q1 = trace[this_node]
        complex_t qx1_q = q1[0].q
        complex_t qy1_q = q1[1].q

        complex_t qx2_q, qy2_q

    if ltree is not None:
        # For non-symmetric traces we have some special checks
        # to do on trees which couldn't be reached from the
        # other dependency trees. Note these are only performed
        # on the left tree; see TraceForest._add_backwards_nonsymm_trees
        # for details.
        if symmetric or (not tree.do_nonsymm_reverse and not tree.do_inv_transform):
            qx2_q = transform_q(tree.left_abcd_x, qx1_q, tree.nr, ltree.nr)
            qy2_q = transform_q(tree.left_abcd_y, qy1_q, tree.nr, ltree.nr)
        elif tree.do_inv_transform:
            # Can't reach tree directly but there is a coupling from ltree.node
            # to tree.node so apply the inverse abcd law to get correct q
            qx2_q = inv_transform_q(tree.left_abcd_x, qx1_q, tree.nr, ltree.nr)
            qy2_q = inv_transform_q(tree.left_abcd_y, qy1_q, tree.nr, ltree.nr)
        else:
            # Really is no way to get to the node (no coupling from ltree.node to
            # tree.node) so only option now is to reverse q for ltree node entry
            qx2_q = -conj(qx1_q)
            qy2_q = -conj(qy1_q)

        qx2 = BeamParam(q=qx2_q, wavelength=lambda0, nr=ltree.nr)
        qy2 = BeamParam(q=qy2_q, wavelength=lambda0, nr=ltree.nr)

        trace[left_node] = qx2, qy2
        if symmetric:
            trace[left_node.opposite] = qx2.reverse(), qy2.reverse()

        propagate(ltree, trace, lambda0, symmetric)

    if rtree is not None:
        qx2_q = transform_q(tree.right_abcd_x, qx1_q, tree.nr, rtree.nr)
        qy2_q = transform_q(tree.right_abcd_y, qy1_q, tree.nr, rtree.nr)

        qx2 = BeamParam(q=qx2_q, wavelength=lambda0, nr=rtree.nr)
        qy2 = BeamParam(q=qy2_q, wavelength=lambda0, nr=rtree.nr)

        trace[right_node] = qx2, qy2
        if symmetric:
            trace[right_node.opposite] = qx2.reverse(), qy2.reverse()

        propagate(rtree, trace, lambda0, symmetric)


cdef bint is_surface_refl(comp, parent_node, sub_node):
    # TODO (sjr) Want to move these imports to module level but we are
    #            losing the war against circular dependencies right now
    from finesse.components.general import InteractionType
    from finesse.components.surface import Surface

    return (
        isinstance(comp, Surface) and
        comp.interaction_type(parent_node, sub_node) == InteractionType.REFLECTION
    )



"""Base solution interface.

Solution classes contain the output from a Finesse simulation, and convenience methods for
accessing, plotting and serialising that output.

Solutions intentionally do not contain references to the model that produced its results. This is so
that the solution can be serialised without requiring the model that produced it itself be
serialisable.
"""
from collections.abc import Iterable
from functools import reduce
from finesse.tree cimport TreeNode
from finesse.solutions.array import ArraySolutionSet
from finesse.solutions.array import ArraySolution

cdef class ParameterChangingTreeNode(TreeNode):
    def __init__(self, name, parent=None):
        super().__init__(name, parent)
        self.parameters_changing = ()

    def get_all_parameters_changing(self):
        return reduce(
            lambda a, b: (*a, *b.parameters_changing),
            self.get_all_children(),
            (*self.parameters_changing,),
        )


cdef class BaseSolution(ParameterChangingTreeNode):
    def __init__(self, name, parent=None):
        super().__init__(name, parent)

    def __str__(self):
        def fn_name(child):
            if type(child) is not BaseSolution:
                r = f" - {child.__class__.__name__}"
            else:
                r = ""
            return child.name + r

        # Override default show_max_children to 3 so large nested solutions do not print huge trees.
        return self.draw_tree(fn_name, title="Solution Tree", show_max_children=3)

    def __getitem__(self, key):
        if isinstance(key, (slice, int)):
            return self.children[key]
        elif isinstance(key, str) or not isinstance(key, Iterable):
            rtn = tuple(child for child in self.children if child.name == str(key))
            if len(rtn) == 1:
                return rtn[0]
            else:
                # return specific set for certain solutions
                if isinstance(rtn[0], ArraySolution):
                    return ArraySolutionSet(rtn)
                else:
                    return rtn
        else:  # We're an iterable that isn't a string
            if len(key) == 1:
                return self[key[0]]
            elif isinstance(key[0], slice):
                raise NotImplementedError(
                    "Solution indexing only supports slices at the end of a key"
                )
            else:
                return self[key[0]][key[1:]]

    def __repr__(self):
        return f"<{self.__class__.__name__} of {self.get_path()} @ {hex(id(self))} children={len(self.children)}>"



from finesse.solutions.base import BaseSolution
from finesse.solutions.base cimport BaseSolution

from cpython.ref cimport PyObject

import numpy as np
cimport numpy as np


cdef class ArraySolution(BaseSolution):
    cdef:
        readonly np.ndarray _outputs
        readonly np.dtype _dtype
        readonly int _axes
        readonly int _num
        public list _units
        readonly tuple x
        readonly tuple params
        readonly tuple detectors
        readonly bint masked
        PyObject** workspaces
        Py_ssize_t num_workspaces

    cpdef int update(self, int index, bint mask)



cimport numpy as np
import numpy as np
cimport cython

from finesse.cymath.homs cimport in_mask
from finesse.detectors.general import Detector


cdef class OutputInformation:
    def __init__(self, name, nodes, dtype, unit, shape, label, needs_fields, needs_trace):
        if shape is None:
            shape = ()
        self.name = name
        self.__nodes = nodes
        self.__dtype = dtype
        self.__dtype_shape = shape
        self.__unit = unit
        self.__label = label
        self.__needs_fields = needs_fields
        self.__needs_trace = needs_trace

    def _set_dtype(self, dtype):
        self.__dtype = dtype

    @property
    def needs_fields(self):
        """Flag indicating whether the detector requires light fields (i.e. solving
        of the interferometer matrix).
        """
        return self.__needs_fields

    @property
    def needs_trace(self):
        """Flag indicating whether the detector requires beam traces."""
        return self.__needs_trace

    @property
    def nodes(self):
        """The nodes this detector observes.

        :getter: Returns the detected node.
        """
        return self.__nodes

    @property
    def dtype(self):
        return self.__dtype

    @property
    def dtype_shape(self):
        return self.__dtype_shape

    def _update_dtype_shape(self, shape):
        """Only to be used internally by detectors like Cameras
        for updating the shape if resolution is changed."""
        self.__dtype_shape = shape

    @property
    def dtype_size(self):
        """Size of the output in terms of number of elements.

        This is typically unity as most detectors return a single
        value via their output functions.

        Equivalent to the product of :attr:`.Detector.dtype_shape`.
        """
        return int(np.prod(self.dtype_shape))

    @property
    def unit(self):
        return self.__unit

    @property
    def label(self):
        return self.__label


cdef class OutputFuncWrapper:
    """
    Helper class for wrapping a C fill function that
    can be referenced from Python by objects. This
    allows a direct C call to the function from other
    cdef functions.

    Examples
    --------
    Create a C function then wrap it using this class:

    >>> cdef void c_output(DetectorWorkspace ptr_ws):
    >>>    cdef PDWorkspace ws = <PDWorkspace>ptr_ws
    >>>    ...
    >>>
    >>> fill = OutputFuncWrapper.make_from_ptr(c_fill)
    """
    def __cinit__(self):
       self.func = NULL

    def __call__(self, ws):
        return self.func(ws)

    @staticmethod
    cdef OutputFuncWrapper make_from_ptr(fptr_c_output f):
        cdef OutputFuncWrapper out = OutputFuncWrapper()
        out.func = f
        return out


cdef class DetectorWorkspace(ElementWorkspace):
    def __init__(self, object owner, object sim, object values=None, *, OutputInformation oinfo=None):
        super().__init__(owner, values)
        self.sim = sim

        if oinfo is None and isinstance(owner, Detector):
            oinfo = owner.output_information
        elif oinfo is None:
            raise ValueError("output information object must be provided if owner is not a detector")

        self.oinfo = oinfo
        self.block_size = oinfo.dtype_size
        self.ignore_sim_mask = False

    cpdef get_output(self):
        self.update_parameter_values()
        if self.fn_c is not None:
            return self.fn_c.func(self)
        elif self.fn_py is not None:
            return self.fn_py(self)
        else:
            return None

    def set_output_fn(self, callback):
        if type(callback) is OutputFuncWrapper:
            self.fn_c = callback
        elif callable(callback):
            self.fn_py = callback
        else:
            raise ValueError(f"Callback function {callback} wasn't of type OutputFuncWrapper or a callable object")


cdef class MaskedDetectorWorkspace(DetectorWorkspace):
    """Specialised workspace for detectors which support masking of modes.

    This workspace provides attributes that are exposed to both C and Python. The
    sections below detail how to use these for some workspace instance ``ws`` which
    inherits from ``MaskedDetectorWorkspace``.

    .. rubric:: Using via Python

    The ``unmasked_indices_arr`` attribute is a :class:`numpy.ndarray`, of dtype
    ``np.intp``, which contains the indices of modes which are not masked. One
    may then simply loop over this array of indices to access the corresponding field
    indices, e.g::

        for k in ws.unmasked_indices_arr:
            # Do something with k, e.g. get field at 0 Hz freq. offset
            # at the given node for the mode index k:
            a_0k = ws.sim.carrier.get_out_fast(ws.dc_node_id, 0, k)
            # use a_0k for some calculation ...

    .. rubric:: Using via Cython

    This workspace also provides a ``unmasked_mode_indices`` pointer (only accessible
    from other Cython code) which corresponds to the data of the ``unmasked_indices_arr``
    NumPy array described above. The attribute ``num_unmasked_homs`` is the size of
    this array; i.e. the number of modes which are not masked.

    One may then write an optimised loop from ``[0, num_unmasked_homs)``, e.g::

        cdef Py_ssize_t i, k
        cdef complex_t a_0k
        for i in range(ws.num_unmasked_homs):
            k = ws.unmasked_mode_indices[i]
            # Do something with k, e.g. get field at 0 Hz freq. offset
            # at the given node for the mode index k:
            a_0k = ws.sim.carrier.get_out_fast(ws.dc_node_id, 0, k)
            # use a_0k for some calculation ...

    where each ``k`` is then the index of the mode at position ``i`` in the unmasked
    indices array.

    .. note::

        If the detector mask is empty (i.e. no modes are being masked) then
        ``unmasked_indices_arr`` (and, correspondingly, ``unmasked_mode_indices``)
        will simply be an array from ``[0, Nhoms)`` where ``Nhoms`` is the total
        number of modes in the simulation.
    """
    def __init__(self, owner, CarrierSignalMatrixSimulation sim, values=None, *, oinfo=None):
        super().__init__(owner, sim, values, oinfo=oinfo)

        self.unmasked_mode_indices = NULL
        self.has_mask = False

        self.setup_mask()

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    cdef int setup_mask(self) except -1:
        cdef:
            int[:, ::1] mask
            Py_ssize_t N_mask
            Py_ssize_t i
            Py_ssize_t count = 0

            int n, m

        # Check if owner has the right attribute, as some non-masked detectors also own masked
        # detector workspaces, and will otherwise throw an error here
        self.has_mask = getattr(self.owner, "has_mask", False)

        cdef np.ndarray[Py_ssize_t, ndim=1, mode="c"] unmasked
        if not self.has_mask: # no mask so just copy across all mode indices
            self.num_unmasked_HOMs = self.sim.model_data.num_HOMs
            unmasked = np.arange(self.num_unmasked_HOMs, dtype=np.intp)
            self.unmasked_indices_arr = unmasked
            self.unmasked_mode_indices = &unmasked[0]
        else:
            mask = self.owner.mask
            N_mask = mask.shape[0]
            if N_mask >= self.sim.model_data.num_HOMs:
                raise RuntimeError(
                    f"Error in detector {self.owner} (Output `{self.oinfo.name}`):\n"
                    f"    Length of mask array ({N_mask}) greater than or equal to "
                    f"number of modes ({self.sim.model_data.num_HOMs})"
                )

            self.num_unmasked_HOMs = self.sim.model_data.num_HOMs - N_mask
            unmasked = np.arange(self.num_unmasked_HOMs, dtype=np.intp)
            self.unmasked_indices_arr = unmasked
            self.unmasked_mode_indices = &unmasked[0]

            for i in range(self.sim.model_data.num_HOMs):
                n = self.sim.model_data.homs_view[i][0]
                m = self.sim.model_data.homs_view[i][1]

                if not in_mask(n, m, mask):
                    if count >= self.num_unmasked_HOMs:
                        raise RuntimeError(
                            f"Error in detector {self.owner.name}:\n"
                            f"    One or more modes in the mask array do not exist in "
                            "the model."
                        )
                    self.unmasked_mode_indices[count] = i
                    count += 1

        return 0

    cdef bint hom_in_modes(self, Py_ssize_t hom_idx):
        """Check whether a HOM index is in the non-masked modes of the detector."""
        if not self.has_mask:
            return True

        cdef Py_ssize_t i
        for i in range(self.num_unmasked_HOMs):
            if self.unmasked_mode_indices[i] == hom_idx:
                return True

        return False



# cython profile=True

import logging

cimport numpy as np
import numpy as np
import cython
from libc.stdlib cimport free, malloc
from cpython.ref cimport PyObject
from finesse.cymath cimport complex_t

from finesse.parameter cimport Parameter
from finesse.simulations.basematrix cimport CarrierSignalMatrixSimulation
from finesse.solutions.array import ArraySolution
from finesse.solutions.array cimport ArraySolution
from finesse.detectors.workspace import DetectorWorkspace
from finesse.detectors.workspace cimport DetectorWorkspace

LOGGER = logging.getLogger(__name__)


ctypedef object (*fptr_callback)(CarrierSignalMatrixSimulation)


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cpdef run_fsig_sweep(
        CarrierSignalMatrixSimulation sim,
        double[::1] axis,
        long[::1] input_rhs_indices,
        long[::1] output_rhs_indices,
        complex_t[:,:,::1] out,
        bint compute_open_loop,
        tuple fsig_independant_outputs = None,
        tuple fsig_dependant_outputs = None
    ):
    """ Runs a simulation to sweep over a signal frequency axis.
    It does this in an optimised way for multiple inputs and outputs.
    It does not use detectors to compute outputs, it will just solve
    the matrix and return transfer functions between nodes. This is
    so it can be used internally for computing TFs without need to add
    detectors everywhere which the user has not specified.
    """
    cdef:
        int Na = len(axis)
        int Ni = len(input_rhs_indices)
        int No = len(output_rhs_indices)
        int i, o, j
        complex_t denom
        Parameter f = sim.model.fsig.f
        bint cast_out = False
        DetectorWorkspace dws
        dict other_outputs = None

    if out is None:
        out = np.zeros((Na, Ni, No), dtype=np.complex128)
        cast_out = True
    else:
        assert(out.shape[0] == Ni)
        assert(out.shape[1] == Na)
        assert(out.shape[2] == No)

    if fsig_independant_outputs or fsig_dependant_outputs:
        other_outputs = {}

    # We'll be making our own RHS inputs for this simulation
    sim.signal.manual_rhs = True
    cdef double ifsig = sim.model_data.fsig

    for j in range(Na):
        f.set_double_value(axis[j])
        sim.model_data.fsig = axis[j]
        sim.signal.refill()
        # For each output that is fsig independant get and store the output
        if fsig_independant_outputs:
            for dws in fsig_independant_outputs:
                other_outputs[dws.oinfo.name] = dws.get_output()

        for i in range(Ni):
            sim.signal.clear_rhs()
            sim.signal.set_source_fast_2(
                input_rhs_indices[i], 1
            )
            sim.signal.solve()
            if not compute_open_loop:
                for o in range(No):
                    out[j][i][o] = sim.signal.out_view[output_rhs_indices[o]]
            else:
                for o in range(No):
                    out[j][i][o] = sim.signal.out_view[output_rhs_indices[o]]
                    if input_rhs_indices[i] == output_rhs_indices[o]:
                        out[j][i][o] -= 1 # remove injected signal
                        out[j][i][o] = out[j][i][o]/(1+out[j][i][o])
                    else:
                        # We can divide out the 1/(1-H) closed loop behaviours by
                        # using the coupling computed back into the same input node
                        denom = sim.signal.out_view[input_rhs_indices[i]]
                        if denom.real == denom.imag == 0:
                            out[j][i][o] = 0
                        else:
                            out[j][i][o] /= denom

    sim.signal.manual_rhs = False
    sim.model_data.fsig = ifsig
    f.set_double_value(ifsig)

    if other_outputs is not None:
        if cast_out:
            return np.array(out), other_outputs
        else:
            return out, other_outputs
    else:
        if cast_out:
            return np.array(out)
        else:
            return out


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
def run_axes_scan(
        object state,
        tuple axes,
        tuple params,
        double[:] offsets,
        tuple out_shape,
        ArraySolution solution,
        object pre_step,
        object post_step,
    ):
    cdef:
        CarrierSignalMatrixSimulation sim = state.sim
        int Np = len(params)
        int No = len(offsets)
        int Na = len(axes)
        int Nos = len(out_shape)
        int i, step
        np.ndarray[double, ndim=1, mode='c'] narr
        double** ptr_axes = NULL
        int* ptr_axes_len = NULL
        PyObject** ptr_params = NULL
        bint mask_this_point = False
        int[::1] IDX = np.zeros(Nos, dtype=np.int32)
        int Ntot

        Parameter param

    if(not (Np == No == Na == Nos)):
        raise Exception("Param, offsets, axes, and out_shape lengths are not the same")

    for p in params:
        if p.datatype is not float:
            raise Exception(f"Can only vary parameters with datatype float, not {p.full_name} with {p.datatype}")

    try:
        # Can't have a memory view of typed ndarray apparently.
        # So here we check the axes are double c-continuous and
        # then save the double pointer
        ptr_axes = <double**> malloc(Na*sizeof(double*))
        if not ptr_axes:
            raise MemoryError()

        ptr_axes_len = <int*> malloc(Na * sizeof(int))
        if not ptr_axes_len:
            raise MemoryError()

        for i in range(Na):
            narr = <np.ndarray[double, ndim=1, mode='c']?> axes[i]
            if narr.size != out_shape[i]:
                raise Exception(f"Out shape[{i}]={out_shape[i]} is not the correct size for the axes[i]={narr.size}")

            ptr_axes[i] = &narr[0]
            ptr_axes_len[i] = narr.size

        # Then to get around some annoying python referencing and issues
        # with accessing cdefs of extension types in a memory view we
        # make an array of PyObjects
        ptr_params = <PyObject**> malloc(Np*sizeof(PyObject*))
        if not ptr_axes:
            raise MemoryError()
        for i in range(Np):
            ptr_params[i] = <PyObject*>(<Parameter?>params[i])

        Ntot = np.prod(out_shape)

        # Now iterate over the all the axes
        #for step in range(Ntot):
        #for idx in np.ndindex(*out_shape):
        for step in range(Ntot):
            # ------------------------------------------------------
            # PRE STEP
            # ------------------------------------------------------
            if pre_step is not None:
                pre_step._do(state)

            for i in range(Np):
                (<Parameter>ptr_params[i]).set_double_value(ptr_axes[i][IDX[i]] + offsets[i])
            # ------------------------------------------------------
            # DO STEP
            # ------------------------------------------------------
            mask_this_point = sim.run_carrier()

            if not mask_this_point and sim.signal:
                sim.run_signal()
            # ------------------------------------------------------
            # POST STEP
            # ------------------------------------------------------
            if mask_this_point:
                values_str = ""
                for i in range(Np):
                    param = <Parameter>ptr_params[i]
                    if param.__units is not None:
                        param_units = " " + param.__units
                    else:
                        param_units = ""

                    values_str += (
                        param.__full_name
                        + " = "
                        + str(ptr_axes[i][IDX[i]] + offsets[i])
                        + param_units
                    )
                    if i != Np - 1:
                        values_str += ", "

                LOGGER.error("Masking simulation outputs at: %s", values_str)

            solution.update(step, mask_this_point)

            if post_step is not None:
                post_step._do(state)

            # ------------------------------------------------------

            # Increment the index vector
            for i in range(No):
                i = Nos-i-1
                IDX[i] += 1
                if IDX[i] >= ptr_axes_len[i]:
                    IDX[i] = 0
                else:
                    break

    finally:
        if ptr_axes != NULL: free(ptr_axes)
        if ptr_axes_len != NULL: free(ptr_axes_len)
        if ptr_params != NULL: free(ptr_params)



from finesse.cymath cimport complex_t
from finesse.cymath.complex cimport conj, cexp, ceq, creal, cimag, cabs
from finesse.cymath.math cimport fabs, radians, degrees
from finesse.cymath.gaussbeam cimport bp_gouy, transform_q
from finesse.cmatrix cimport SubCCSView, SubCCSView1DArray, SubCCSView2DArray
from finesse.knm cimport KnmMatrix
from finesse.simulations.base cimport ModelData, NodeBeamParam
from finesse.frequency cimport frequency_info_t
from finesse.simulations.basematrix cimport CarrierSignalMatrixSimulation, MatrixSystemSolver
from finesse.symbols import Symbol

import numpy as np
cimport numpy as np
cimport cython

from cpython.ref cimport PyObject, Py_XINCREF, Py_XDECREF
from libc.string cimport strcmp, memcpy
from libc.stdlib cimport free, calloc


cdef extern from "math.h":
    double sin(double)
    double cos(double)


ctypedef (double*, double*, double*, double*) ptr_tuple_4


cdef extern from "constants.h":
    long double PI
    double C_LIGHT
    double DEG2RAD


cdef class SpaceOpticalConnections:
    def __cinit__(self, MatrixSystemSolver mtx):
        # Only 1D arrays of views as spaces don't
        # couple frequencies together.
        Nf = mtx.optical_frequencies.size
        self.P1i_P2o = SubCCSView1DArray(Nf)
        self.P2i_P1o = SubCCSView1DArray(Nf)
        self.opt_ptrs.P1i_P2o = <PyObject**>self.P1i_P2o.views
        self.opt_ptrs.P2i_P1o = <PyObject**>self.P2i_P1o.views


cdef class SpaceSignalConnections(SpaceOpticalConnections):
    def __cinit__(self, MatrixSystemSolver mtx):
        # Only 1D arrays of views as spaces don't
        # couple frequencies together.
        if not mtx.is_signal_matrix:
            raise Exception("Signal simulation not enabled")
        Nf = mtx.optical_frequencies.size
        self.SIGAMP_P1o = SubCCSView2DArray(1, Nf)
        self.SIGAMP_P2o = SubCCSView2DArray(1, Nf)
        self.SIGPHS_P1o = SubCCSView2DArray(1, Nf)
        self.SIGPHS_P2o = SubCCSView2DArray(1, Nf)
        self.H_P1o = SubCCSView2DArray(1, Nf)
        self.H_P2o = SubCCSView2DArray(1, Nf)

        self.sig_ptrs.SIGAMP_P1o = <PyObject***>self.SIGAMP_P1o.views
        self.sig_ptrs.SIGAMP_P2o = <PyObject***>self.SIGAMP_P2o.views
        self.sig_ptrs.SIGPHS_P1o = <PyObject***>self.SIGPHS_P1o.views
        self.sig_ptrs.SIGPHS_P2o = <PyObject***>self.SIGPHS_P2o.views
        self.sig_ptrs.H_P1o = <PyObject***>self.H_P1o.views
        self.sig_ptrs.H_P2o = <PyObject***>self.H_P2o.views


cdef class SpaceValues(BaseCValues):
    def __init__(self):
        cdef ptr_tuple_4 ptr = (&self.L, &self.nr, &self.gouy_x, &self.gouy_y)
        cdef tuple params = ("L","nr","gouy_x","gouy_y")
        self.setup(params, sizeof(ptr), <double**>&ptr)


cdef class SpaceWorkspace(ConnectorWorkspace):
    def __init__(self, object owner, CarrierSignalMatrixSimulation sim, bint refill):
        super().__init__(
            owner, sim, refill, refill,
            SpaceOpticalConnections(sim.carrier),
            SpaceSignalConnections(sim.signal) if sim.signal else None,
            SpaceValues()
        )
        # Here we cast connections and values to a non-specific object
        # so the fill functions don't have to and cython can optimise better
        self.sco = self.carrier.connections
        self.scs = self.signal.connections if sim.signal else None
        self.sv = self.values

        self.P1i_id = sim.carrier.node_id(owner.p1.i)
        self.P1o_id = sim.carrier.node_id(owner.p1.o)
        self.P2i_id = sim.carrier.node_id(owner.p2.i)
        self.P2o_id = sim.carrier.node_id(owner.p2.o)

        if sim.signal:
            # If we have a signal simulation then we need to cache some indicies
            # for grabbing data when filling
            self.car_p1o_rhs_idx = sim.carrier.get_node_info(owner.p1.o)['rhs_index']
            self.car_p2o_rhs_idx = sim.carrier.get_node_info(owner.p2.o)['rhs_index']
            self.car_p_num_hom =   sim.carrier.get_node_info(owner.p1.o)['nhoms']
            self.strain_signal_enabled = owner.h.i.full_name in sim.signal.nodes
            self.phase_signal_enabled = owner.phs.i.full_name in sim.signal.nodes
            self.amp_signal_enabled = owner.amp.i.full_name in sim.signal.nodes

        self.couplings = <complex_t*> calloc(sim.model_data.num_HOMs, sizeof(complex_t))
        self.sym_abcd_B = NULL

    def __dealloc__(self):
        if self.couplings:
            free(self.couplings)

        cy_expr_free(self.sym_abcd_B)

    cpdef compile_abcd_cy_exprs(self):
        cdef:
            object space = self.owner
            cdef object[:, ::1] M_sym = list(space._abcd_matrices.values())[0][0]

        # NOTE (sjr) Only element B of a Space ABCD matrix can possibly change
        if isinstance(M_sym[0][1], Symbol):
            ch_sym = M_sym[0][1].expand_symbols().eval(keep_changing_symbols=True)
            if isinstance(ch_sym, Symbol):
                self.sym_abcd_B = cy_expr_new()
                cy_expr_init(self.sym_abcd_B, ch_sym)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    cpdef update_parameter_values(self):
        ConnectorWorkspace.update_parameter_values(self)

        if self.sym_abcd_B != NULL:
            self.abcd[0][1] = cy_expr_eval(self.sym_abcd_B)


space_carrier_fill = FillFuncWrapper.make_from_ptr(c_space_carrier_fill)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cdef object c_space_carrier_fill(ConnectorWorkspace cws):
    r"""
    Fills the sub-matrix of the interferometer matrix held by `sim`, corresponding
    to the `space` component.

    The space component propagates a light field over a length :math:`L` with index of refraction
    :math:`n_r`. The product :math:`n_rL` is, by definition in Finesse, always a multiple of the default
    laser wavelength :math:`\lambda_0`. This defines a *macroscopic length*.

    .. _fig_space_couplings:
    .. figure:: ../images/space.*
        :align: center

        Field couplings for a space.

    The propagation only affects the phase of the field (in the plane-wave picture):

    .. math::
        s_1 = s_2 = \exp{\left(-i(\omega_0 + \Delta\omega) n_rL / c\right)}
        = \exp{\left(-i\Delta\omega n_rL / c\right)},

    where :math:`\exp{\left(-i\omega_0 n_rL / c\right)} = 1` following from the definition of macroscopic lengths
    above. The quantity :math:`\Delta\omega` is the offset to the default (angular) frequency :math:`\omega`.

    In the modal picture, Gouy phases are accumulated over space components. This results in the coupling equation
    including these phase terms such that each field of a different mode order has a different phase accumulation:

    .. math::
        s_{1, \mathrm{nm}} = s_{2, \mathrm{nm}} = \exp{\left(-i(\Delta\omega n_rL / c + n\psi_x + m\psi_y)\right)},

    where :math:`n, m` are the mode indices and :math:`\psi_x, \psi_y` are the Gouy phases of the space in tangential
    and sagittal planes, respectively. Note that if the flag `zero_tem00_gouy` is false, then :math:`n \rightarrow n + 1/2`
    and :math:`m \rightarrow m + 1/2`.

    Parameters
    ----------
    """
    cdef:
        SpaceWorkspace ws = <SpaceWorkspace>cws
        space_optical_connections *conns = <space_optical_connections*>&ws.sco.opt_ptrs
        double f0 = ws.sim.model_data.f0

        double pre_factor = 2 * PI * ws.sv.nr * ws.sv.L / C_LIGHT
        double epsilon0_c = ws.sim.model_data.EPSILON0_C
        double f_car
        double fs
        double phi
        complex_t factor

        double gouy_x = radians(ws.sv.gouy_x)
        double gouy_y = radians(ws.sv.gouy_y)

        Py_ssize_t field_idx, size
        int n, m, i
        double ni, mi, gouy
        frequency_info_t *frequencies
        frequency_info_t *freq

    frequencies = ws.sim.carrier.optical_frequencies.frequency_info
    size = ws.sim.carrier.optical_frequencies.size

    for i in range(size):
        freq = &frequencies[i]
        space_fill_optical_2_optical(
            conns, ws, freq, pre_factor,
            gouy_x, gouy_y,
            ws.sim.model_data.zero_tem00_gouy,
            ws.sim.model_data.homs_view
        )

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cdef inline void space_fill_optical_2_optical(
        space_optical_connections *conn,
        SpaceWorkspace ws,
        frequency_info_t *freq,
        double pre_factor,
        double gouy_x,
        double gouy_y,
        bint zero_tem00_gouy,
        int[:, ::1] homs_view
    ):
    cdef:
        int n, m
        double ni, nm

    phi = pre_factor * freq.f

    for field_idx in range(homs_view.shape[0]):
        n = homs_view[field_idx][0]
        m = homs_view[field_idx][1]

        if zero_tem00_gouy:
            ni = n
            mi = m
        else:
            ni = n + 0.5
            mi = m + 0.5

        gouy = ni * gouy_x + mi * gouy_y
        ws.couplings[field_idx] = cexp(1j * (-phi + gouy))

    # Fill diagonals with the vector
    if conn.P1i_P2o[freq.index]:
        (<SubCCSView>conn.P1i_P2o[freq.index]).fill_negative_zd_2(ws.couplings, 1)
    if conn.P2i_P1o[freq.index]:
        (<SubCCSView>conn.P2i_P1o[freq.index]).fill_negative_zd_2(ws.couplings, 1)


space_signal_fill = FillFuncWrapper.make_from_ptr(c_space_signal_fill)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cdef object c_space_signal_fill(ConnectorWorkspace cws):
    cdef:
        SpaceWorkspace ws = <SpaceWorkspace>cws
        space_optical_connections *oconns = <space_optical_connections*>&ws.scs.opt_ptrs
        space_signal_connections *sconns = <space_signal_connections*>&ws.scs.sig_ptrs
        double f0 = ws.sim.model_data.f0

        double pre_factor = 2 * PI * ws.sv.nr * ws.sv.L / C_LIGHT
        double epsilon0_c = ws.sim.model_data.EPSILON0_C
        double f_car
        double fs
        double phi
        complex_t factor

        double gouy_x = radians(ws.sv.gouy_x)
        double gouy_y = radians(ws.sv.gouy_y)

        Py_ssize_t field_idx, size
        int n, m, i
        double ni, mi, gouy
        frequency_info_t *frequencies

    frequencies = ws.sim.signal.optical_frequencies.frequency_info
    size = ws.sim.signal.optical_frequencies.size

    for i in range(size): # for each optical signal sideband
        space_fill_optical_2_optical(
            oconns, ws, &frequencies[i], pre_factor,
            gouy_x, gouy_y,
            ws.sim.model_data.zero_tem00_gouy,
            ws.sim.model_data.homs_view
        )

    if ws.strain_signal_enabled:
        for i in range(size): # for each optical signal sideband
            strain_signal_fill(sconns, ws, &frequencies[i])

    if ws.phase_signal_enabled:
        for i in range(size): # for each optical signal sideband
            phase_signal_fill(sconns, ws, &frequencies[i])

    if ws.amp_signal_enabled:
        for i in range(size): # for each optical signal sideband
            amplitude_signal_fill(sconns, ws, &frequencies[i])


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cdef inline void phase_signal_fill(
    space_signal_connections *sconns,
    SpaceWorkspace ws,
    frequency_info_t *freq
):
    cdef:
        const complex_t* c_p1_o = NULL
        const complex_t* c_p2_o = NULL
        complex_t z = -1j * 0.5 # Just some pure phase modulation on the space outputs
        # Minus sign here because we model phase propagation as exp(-1j*(phi+A sin(Omega*t)))

    # Here we use the outgoing carrier field at each port.
    # This ensures that we include any gouy phase terms, etc.
    # that would otherwise need to be applied here as well,
    # compared to using the incoming fields.
    c_p1_o = &ws.sim.carrier.out_view[ws.sim.carrier.field_fast_2(
        ws.car_p1o_rhs_idx,
        ws.car_p_num_hom,
        freq.audio_carrier_index,
        0 # Select zeroth element, HOM storage is continuous in memory
    )]
    c_p2_o = &ws.sim.carrier.out_view[ws.sim.carrier.field_fast_2(
        ws.car_p2o_rhs_idx,
        ws.car_p_num_hom,
        freq.audio_carrier_index,
        0 # Select zeroth element, HOM storage is continuous in memory
    )]

    if sconns.SIGPHS_P1o[0][freq.index]:
        # Assume H only has a single frequency as input...
        (<SubCCSView>sconns.SIGPHS_P1o[0][freq.index]).fill_negative_za_zd_2 (
            z, c_p1_o, 1 # 1D contiguous array from outview
        )

    if sconns.SIGPHS_P2o[0][freq.index]:
        # Assume H only has a single frequency as input...
        (<SubCCSView>sconns.SIGPHS_P2o[0][freq.index]).fill_negative_za_zd_2 (
            z, c_p2_o, 1 # 1D contiguous array from outview
        )


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cdef inline void amplitude_signal_fill(
    space_signal_connections *sconns,
    SpaceWorkspace ws,
    frequency_info_t *freq
):
    cdef:
        const complex_t* c_p1_o = NULL
        const complex_t* c_p2_o = NULL
        # Just some pure amplitude modulate whatever
        # is coming out of the space ports
        complex_t z = 0.5

    # Here we use the outgoing carrier field at each port.
    # This ensures that we include any gouy phase terms, etc.
    # that would otherwise need to be applied here as well,
    # compared to using the incoming fields.
    c_p1_o = &ws.sim.carrier.out_view[ws.sim.carrier.field_fast_2(
        ws.car_p1o_rhs_idx,
        ws.car_p_num_hom,
        freq.audio_carrier_index,
        0 # Select zeroth element, HOM storage is continuous in memory
    )]
    c_p2_o = &ws.sim.carrier.out_view[ws.sim.carrier.field_fast_2(
        ws.car_p2o_rhs_idx,
        ws.car_p_num_hom,
        freq.audio_carrier_index,
        0 # Select zeroth element, HOM storage is continuous in memory
    )]

    if sconns.SIGAMP_P1o[0][freq.index]:
        # Assume H only has a single frequency as input...
        (<SubCCSView>sconns.SIGAMP_P1o[0][freq.index]).fill_negative_za_zd_2 (
            z, c_p1_o, 1 # 1D contiguous array from outview
        )

    if sconns.SIGAMP_P2o[0][freq.index]:
        # Assume H only has a single frequency as input...
        (<SubCCSView>sconns.SIGAMP_P2o[0][freq.index]).fill_negative_za_zd_2 (
            z, c_p2_o, 1 # 1D contiguous array from outview
        )


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cdef void strain_signal_fill(
    space_signal_connections *sconns,
    SpaceWorkspace ws,
    frequency_info_t *freq
):
    cdef:
        double w_g, m_g, phi_g, w_0
        const complex_t* c_p1_o = NULL
        const complex_t* c_p2_o = NULL
        complex_t z

    # Reference for strain signal over a space...
    # Interferometer responses to gravitational waves:
    # Comparing Finesse simulations and analytical solutions
    # Charlotte Bond, Daniel Brown and Andreas Freise
    # LIGO DCC: T1300190
    # https://arxiv.org/pdf/1306.6752.pdf
    w_g = 2 * PI * ws.sim.model_data.fsig  # GW frequency [rad/s]
    # What isn't really noted in the document is signals around multiple carrier
    # fields, but this is really just a scaling of w0
    w_0 = 2 * PI * (ws.sim.model_data.f0 + freq.f_car[0])
    # 1/2 factor from Eq.7
    m_g = - 0.5 * w_0/w_g * sin(w_g * ws.sv.L * ws.sv.nr / 2 / C_LIGHT) # GW signal per strain [1/h] Eq.14
    phi_sb = - w_g * ws.sv.L * ws.sv.nr / 2 / C_LIGHT # Eq.15
    # audio order here takes care of minus sign options in Eq.17
    z = 1j*m_g * cexp(1j * phi_sb * freq.audio_order) # i factor from pi/2 in Eq.17
    # Carrier phase term in Eq.17 is included already using the out
    # going field at the ports below

    # Here we use the outgoing carrier field at each port.
    # This ensures that we include any gouy phase terms, etc.
    # that would otherwise need to be applied here as well,
    # compared to using the incoming fields.
    c_p1_o = &ws.sim.carrier.out_view[ws.sim.carrier.field_fast_2(
        ws.car_p1o_rhs_idx,
        ws.car_p_num_hom,
        freq.audio_carrier_index,
        0 # Select zeroth element, HOM storage is continuous in memory
    )]
    c_p2_o = &ws.sim.carrier.out_view[ws.sim.carrier.field_fast_2(
        ws.car_p2o_rhs_idx,
        ws.car_p_num_hom,
        freq.audio_carrier_index,
        0 # Select zeroth element, HOM storage is continuous in memory
    )]

    if sconns.H_P1o[0][freq.index]:
        # Assume H only has a single frequency as input...
        (<SubCCSView>sconns.H_P1o[0][freq.index]).fill_negative_za_zd_2 (
            z, c_p1_o, 1 # 1D contiguous array from outview
        )

    if sconns.H_P2o[0][freq.index]:
        # Assume H only has a single frequency as input...
        (<SubCCSView>sconns.H_P2o[0][freq.index]).fill_negative_za_zd_2 (
            z, c_p2_o, 1 # 1D contiguous array from outview
        )


space_set_gouy = GouyFuncWrapper.make_from_ptr(c_set_gouy_phase)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cdef int c_set_gouy_phase(ConnectorWorkspace cws) except -1:
    cdef:
        SpaceWorkspace ws = <SpaceWorkspace>cws
        const NodeBeamParam* q_p1o = &ws.sim.trace[ws.P1i_id]
        const NodeBeamParam* q_p2i = &ws.sim.trace[ws.P2o_id]

        double nr = ws.sv.nr

        complex_t qx_p1o_propagated
        complex_t qy_p1o_propagated

    if not q_p1o.is_changing and not q_p2i.is_changing:
        return 0

    qx_p1o_propagated = transform_q(
        ws.abcd, q_p1o.qx.q, nr, nr
    )
    qy_p1o_propagated = transform_q(
        ws.abcd, q_p1o.qy.q, nr, nr
    )

    # Can't have mismatches across spaces, these must occur at
    # components, so if a mismatch is present then report a bug
    # with an informative error message.
    if not ceq(q_p2i.qx.q, qx_p1o_propagated):
        raise RuntimeError(
            _space_mismatch_bug_report(
                ws,
                ws.P1i_id,
                ws.P2o_id,
                q_p1o.qx.q,
                q_p2i.qx.q,
                qx_p1o_propagated,
                "x"
            )
        )

    if not ceq(q_p2i.qy.q, qy_p1o_propagated):
        raise RuntimeError(
            _space_mismatch_bug_report(
                ws,
                ws.P1i_id,
                ws.P2o_id,
                q_p1o.qy.q,
                q_p2i.qy.q,
                qy_p1o_propagated,
                "y"
            )
        )

    ws.owner.gouy_x = degrees(
        fabs(bp_gouy(&q_p2i.qx) - bp_gouy(&q_p1o.qx))
    )
    ws.owner.gouy_y = degrees(
        fabs(bp_gouy(&q_p2i.qy) - bp_gouy(&q_p1o.qy))
    )

    return 0


cdef _space_mismatch_bug_report(
    SpaceWorkspace ws,
    Py_ssize_t node1_id,
    Py_ssize_t node2_id,
    complex_t q1,
    complex_t q2,
    complex_t q2e, # expected q
    direction, # "x" or "y"
):
    node1_name = list(ws.sim.carrier.nodes.keys())[node1_id]
    node2_name = list(ws.sim.carrier.nodes.keys())[node2_id]

    z_diff = fabs(creal(q2e) - creal(q2))
    zr_diff = fabs(cimag(q2e) - cimag(q2))
    mag_diff = cabs(q2e - q2)

    return (
        f"Bug encountered! Mismatch in space {ws.owner.name} "
        f"from {node1_name} -> {node2_name}:"
        f"\n    STORED q{direction} @ {node2_name} = {q2}"
        f"\n    EXPECTED q{direction} = {q2e}"
        f"\n        given q{direction} @ {node1_name} = {q1},"
        f"\n        with ABCD = {ws.abcd.base.tolist()}"
        f"\n    DIFFERENCES: |z2 - z1| = {z_diff}, |zr2 - zr1| = {zr_diff};"
        f"\n                 |q1 - q2| = {mag_diff}"
    )



import contextlib

import logging

cimport cython

cimport numpy as np
import numpy as np

import networkx as nx

from finesse.cymath cimport complex_t
from finesse.cymath.complex cimport conj
from finesse.cymath.math cimport float_eq
from finesse.cymath.gaussbeam cimport transform_q, inv_transform_q
from finesse.element cimport ElementWorkspace
from finesse.frequency cimport Frequency
from finesse.components.workspace cimport ConnectorCallbacks, ConnectorWorkspace
from finesse.components.node import NodeType

from finesse.components.modal.cavity cimport CavityWorkspace

from cpython.ref cimport PyObject
from libc.stdio cimport printf


cdef extern from "stdlib.h":
    void  free(void* ptr)
    void* malloc(size_t size)
    void* calloc(size_t N, size_t size)


cdef extern from "constants.h":
    long double PI
    double C_LIGHT
    double complex COMPLEX_0


LOGGER = logging.getLogger(__name__)


cdef class ModelData:
    pass



#cython: boundscheck=False, wraparound=False, initializedcheck=False

cimport cython

from finesse.cymath cimport complex_t
from finesse.cymath.gaussbeam cimport transform_q
from finesse.simulations.base cimport NodeBeamParam
from finesse.simulations.basematrix cimport CarrierSignalMatrixSimulation
from finesse.cmatrix cimport SubCCSView1DArray
from finesse.symbols import Symbol

from cpython.ref cimport PyObject

import logging

ctypedef (double*, ) ptr_tuple_1


LOGGER = logging.getLogger(__name__)


cdef class LensValues(BaseCValues):
    def __init__(LensValues self):
        cdef ptr_tuple_1 ptr = (&self.f, )
        cdef tuple params = ("f", )
        self.setup(params, sizeof(ptr), <double**>&ptr)


cdef class LensConnections:
    def __cinit__(self, CarrierSignalMatrixSimulation sim):
        self.P1i_P2o = SubCCSView1DArray(sim.carrier.optical_frequencies.size)
        self.P2i_P1o = SubCCSView1DArray(sim.carrier.optical_frequencies.size)

        self.conn_ptrs.P1i_P2o = <PyObject**>self.P1i_P2o.views
        self.conn_ptrs.P2i_P1o = <PyObject**>self.P2i_P1o.views


cdef class LensWorkspace(KnmConnectorWorkspace):
    def __init__(self, owner, CarrierSignalMatrixSimulation sim, refill):
        super().__init__(
            owner,
            sim,
            refill,
            refill,
            LensConnections(sim),
            None,
            LensValues()
        )
        self.lc = self.carrier.connections
        self.lv = self.values

        self.P1i_id = sim.carrier.node_id(owner.p1.i)
        self.P1o_id = sim.carrier.node_id(owner.p1.o)
        self.P2i_id = sim.carrier.node_id(owner.p2.i)
        self.P2o_id = sim.carrier.node_id(owner.p2.o)

        self.sym_abcd_Cx = NULL
        self.sym_abcd_Cy = NULL

    def __dealloc__(self):
        cy_expr_free(self.sym_abcd_Cx)
        cy_expr_free(self.sym_abcd_Cy)

    cpdef compile_abcd_cy_exprs(self):
        cdef:
            dict abcd_handles = self.owner._abcd_matrices
            tuple keyx = (self.owner.p1.i, self.owner.p2.o, "x")
            tuple keyy = (self.owner.p1.i, self.owner.p2.o, "y")
            cdef object[:, ::1] Mx_sym = abcd_handles[keyx][0]
            cdef object[:, ::1] My_sym = abcd_handles[keyy][0]

        # NOTE (sjr) Only element C of a Lens ABCD matrix can possibly change

        if isinstance(Mx_sym[1][0], Symbol):
            ch_sym = Mx_sym[1][0].expand_symbols().eval(keep_changing_symbols=True)
            if isinstance(ch_sym, Symbol):
                self.sym_abcd_Cx = cy_expr_new()
                cy_expr_init(self.sym_abcd_Cx, ch_sym)

        if isinstance(My_sym[1][0], Symbol):
            ch_sym = My_sym[1][0].expand_symbols().eval(keep_changing_symbols=True)
            if isinstance(ch_sym, Symbol):
                self.sym_abcd_Cy = cy_expr_new()
                cy_expr_init(self.sym_abcd_Cy, ch_sym)

    cpdef update_parameter_values(self):
        ConnectorWorkspace.update_parameter_values(self)

        if self.sym_abcd_Cx != NULL:
            self.abcd_x[1][0] = cy_expr_eval(self.sym_abcd_Cx)

        if self.sym_abcd_Cy != NULL:
            self.abcd_y[1][0] = cy_expr_eval(self.sym_abcd_Cy)

# TODO (sjr) make c_lens_fill function?



"""
"""

import enum
import logging

LOGGER = logging.getLogger(__name__)


@enum.unique
class SpatialType(enum.Enum):
    """The spatial type of the model - i.e. either plane wave or modal based."""

    PLANE = 0
    MODAL = 1



"""List available Finesse elements.

Originally used for #45.
"""

from finesse.element import ModelElement


def inheritors(klass):
    subclasses = set()
    work = [klass]
    while work:
        parent = work.pop()
        for child in parent.__subclasses__():
            if child not in subclasses:
                subclasses.add(child)
                work.append(child)
    return subclasses


MODEL_ELEMENT_CHILDREN = inheritors(ModelElement)


if __name__ == "__main__":
    for cls in sorted(
        MODEL_ELEMENT_CHILDREN, key=lambda cls: (cls.__module__, cls.__name__)
    ):
        print(f"* [ ] `{cls.__module__}.{cls.__name__}`")



from finesse.cymath cimport complex_t
from finesse.cymath.gaussbeam cimport beam_param
from finesse.components.modal.cavity cimport CavityWorkspace
from finesse.detectors.workspace cimport DetectorWorkspace
from finesse.simulations.base cimport NodeBeamParam

### Beam Property Detector ###


cpdef enum BeamProperty:
    SIZE = 0
    WAISTSIZE = 1
    DISTANCE = 2
    RAYLEIGH = 3
    GOUY = 4
    DIVERGENCE = 5
    ROC = 6
    Q = 7
    DEFOCUS = 8

cdef class BPDetectorWorkspace(DetectorWorkspace):
    cdef readonly:
        BeamProperty detecting

        bint is_changing
        double initial
    cdef public:
        bint q_as_bp # Whether to return q as a BeamParam or just complex_t

    cdef:
        const beam_param* q
        double (*compute_func)(const beam_param*)


### Cavity Property Detector ###


cpdef enum CavityProperty:
    LENGTH = 0
    LOSS = 1
    FINESSE = 2
    FSR = 3
    FWHM = 4
    POLE = 5
    TAU = 6
    ABCD = 7
    STABILITY = 8
    RTGOUY = 9
    MODESEP = 10
    RESOLUTION = 11
    EIGENMODE = 12
    SOURCE_SIZE = 13
    SOURCE_WAISTSIZE = 14
    SOURCE_DISTANCE = 15
    SOURCE_RAYLEIGH = 16
    SOURCE_DIVERGENCE = 17
    SOURCE_ROC = 18
    SOURCE_DEFOCUS = 19

cdef class CPDetectorWorkspace(DetectorWorkspace):
    cdef readonly:
        CavityProperty detecting

    cdef:
        # Pointer to the attribute of CavityWorkspace being detected
        double* target

cdef class CPDetectorABCDWorkspace(DetectorWorkspace):
    cdef double[:, ::1] abcd

cdef class CPDetectorModeWorkspace(DetectorWorkspace):
    cdef readonly:
        CavityProperty detecting

        bint is_changing
        double initial
    cdef:
        # Pointer to eigenmode
        const beam_param* q
        # Pointer to flag for whether cavity is stable in relevant plane
        bint* is_stable

        double (*compute_func)(const beam_param*)

    cdef public:
        bint q_as_bp # Whether to return q as a BeamParam or just complex_t


### Astigmatism Detector ###


cdef class AstigmatismDetectorWorkspace(DetectorWorkspace):
    # Pointer to beam parameter entry in sim.trace
    cdef const NodeBeamParam* q

    cdef readonly double initial


### Mode Mismatch Detector ###


cdef class ModeMismatchDetectorWorkspace(DetectorWorkspace):
    cdef readonly:
        int pscale

        bint is_changing
        double initial
    cdef:
        # Pointers to beam parameter entries in sim.trace
        const beam_param* q1
        const beam_param* q2
        const double[:, ::1] abcd


### Accumulated Gouy Phase Detector ###


cdef class GouyDetectorWorkspace(DetectorWorkspace):
    # Array of pointers to gouy_[x|y] attribute of each space workspace
    cdef double** gouy_targets
    cdef Py_ssize_t Nspaces



"""Simulation output"""

import numpy as np
cimport numpy as np
import logging
import cython

from finesse.plotting import Plotter
from finesse.detectors.workspace cimport DetectorWorkspace
from finesse.solutions.base cimport BaseSolution

from cpython.ref cimport PyObject, Py_XINCREF, Py_XDECREF
from libc.stdlib cimport calloc, free

LOGGER = logging.getLogger(__name__)

from collections.abc import Set, Iterable

class ArraySolutionSet(Set):
    """
    A set of Arraysolution. Outputs from mutliple similar ArraySolutions
    can be returned.
    """
    def __init__(self, solutions):
        if not isinstance(solutions, Iterable):
            raise Exception(f"{solutions} is not iterable")
        if not all(isinstance(_, ArraySolution) for _ in solutions):
            raise Exception(f"Solutions {solutions} must all be of type ArraySolution")
        if not all(_.outputs == solutions[0].outputs for _ in solutions[1:]):
            raise Exception(f"Solutions {solutions} must have the same outputs")

        self.solutions = lst = []
        for value in solutions:
            if value not in lst:
                lst.append(value)

    def __iter__(self):
        return iter(self.solutions)

    def __contains__(self, value):
        return value in self.solutions

    def __len__(self):
        return len(self.solutions)

    def __getitem__(self, key):
        return np.vstack(tuple(_[key] for _ in self.solutions))

    @property
    def outputs(self):
        return self.solutions[0].outputs

cdef class ArraySolution(BaseSolution):
    """Holds outputs from running a simulation.

    This is essentially a wrapped up Numpy structured array whose named elements are the names of
    outputs in a model.

    Detectors are stored in the array by their name. So you can use::

        output['detector_name']

    or, if the key has an attribute called `name` (as all Finesse.detectors do)
    it will use that, so using::

        output[ifo.detector]

    will return the same values.

    The underlying storage format is a Numpy structured array. You can select
    runs by::

        output[ifo.detector][a:b:c]

    where `a:b:c` is your slice. Or you can select multiple outputs with::

        output[['det1', 'det2']][a:b:c]

    Parameters
    ----------
    name : str
        Name to give to this analysis

    parents : :class:`.BaseSolution`
        Parent solutions that have preceeded this solution being calculated. Can be
        `None`.

    model : :class:`.Model`
        Model object to get simulation data from

    shape : :class:`numpy.ndarray`
        The shape of the underlying data array. use a single integer for 1D
        outputs, N-dimensional outputs can be specified by using tuples, i.e.
        (10,5,100) for a 3D array with the requested sizes.
    """
    def __cinit__(self, *args, **kwargs):
        self.workspaces = NULL
        self.masked = False

    def __dealloc__(self):
        if self.workspaces != NULL:
            for i in range(self.num_workspaces):
                if self.workspaces[i] != NULL:
                    Py_XDECREF(self.workspaces[i])

            free(self.workspaces)
            self.workspaces = NULL

    def __init__(self, name, parents, detector_workspaces, shape, xs, params):
        cdef Py_ssize_t i

        super().__init__(name, parents)
        self._num = -1
        self._units = []
        self.empty = False  # tree drawing fill circle
        self.x = tuple(xs)
        self.params = tuple(params)
        self.detectors = tuple(ws.owner for ws in detector_workspaces)

        if hasattr(shape, "__getitem__"):
            self._axes = len(shape)
        else:
            self._axes = 1

        dtypes = []

        for ws in detector_workspaces:
            oinfo = ws.oinfo
            dtypes.append((oinfo.name, oinfo.dtype, oinfo.dtype_shape))

        LOGGER.info("Outputting shape=%s dtype=%s", shape, dtypes)

        self._dtype = np.dtype(dtypes)
        self._outputs = np.zeros(shape, dtype=self._dtype)

        # Setup the workspace pointers
        self.num_workspaces = len(detector_workspaces)

        if self.num_workspaces > 0:
            self.workspaces = <PyObject**>calloc(self.num_workspaces, sizeof(PyObject*))
            if not self.workspaces:
                raise MemoryError()

            for i in range(self.num_workspaces):
                oinfo = detector_workspaces[i].oinfo
                self.workspaces[i] = <PyObject*>detector_workspaces[i]
                Py_XINCREF(self.workspaces[i])
                (<DetectorWorkspace?>self.workspaces[i]).flatiter = self._outputs[oinfo.name].flat

    def __getattr__(self, key):
        try:
            if key[0] == 'x':
                idx = int(key[1:])
                if 1 < idx > len(self.x):
                    raise IndexError(f"Can only select x1 to x{len(self.x)}")
                return self.x[idx-1]
            if key[0] == 'p':
                idx = int(key[1:])
                if 1 < idx > len(self.params):
                    raise IndexError(f"Can only select p1 to p{len(self.params)}")
                return self.params[idx-1]
        except ValueError:
            pass

    @property
    def data(self):
        return self._outputs

    @property
    def axes(self):
        return self._axes

    @property
    def entries(self):
        """
        The number of outputs that have been stored in here so far
        """
        return self._num + 1

    @property
    def shape(self):
        return self.data.shape

    @property
    def outputs(self):
        """
        Returns all the outputs that have been stored.
        """
        return self._outputs.dtype.names

    def expand(self, shape):
        """
        Expands the output buffer by `shape` elements. This will be slow if you
        call repeatedly to increase by just a few elements as the whole array
        is copied into a new block of memory.

        Parameters
        ----------
        shape
            Shape of new array to make
        """
        self._outputs = np.append(self._outputs, np.empty(shape, dtype=self._dtype))

    def __getitem__(self, key):
        try:
            if not isinstance(key, str) and isinstance(key, Iterable) and len(key) == 1:
                key = key[0]

            if hasattr(key, "name"):
                key = key.name

            # NOTE: indexing with empty tuple here to handle noxaxis type outputs
            out = self._outputs[key][()]
            if self.masked:
                if out.dtype == object:
                    return np.ma.masked_object(out, np.nan)[()]
                else:
                    return np.ma.masked_invalid(out)[()]
            else:
                return out

        except (ValueError, IndexError, TypeError):
            return super().__getitem__(key)

    @cython.boundscheck(False)
    @cython.initializedcheck(False)
    @cython.wraparound(False)
    cpdef int update(self, int index, bint mask):
        """
        Calling this will compute all detector outputs and add an entry to
        the outputs stored. Calling it multiple times without re-running
        the simulation will result in duplicate entries.

        Parameters
        ----------

        index : (int, ...)
           Index to calculate the outputs for, use tuples of N-size for
           N-dimensional outputs

        mask : boolean
            Sets the `index` of all outputs to `np.nan` if true. This should
            be used when a recoverable invalidity has occurred at some point
            in the scan of a parameter of a simulation - e.g. invalid beam
            tracing due to instabilities.
        """
        cdef:
            Py_ssize_t i
            DetectorWorkspace dws

            # Indices for start and end of blocks for
            # detectors with array-like outputs
            Py_ssize_t block_i, block_f

        self.masked |= mask
        for i in range(self.num_workspaces):
            dws = <DetectorWorkspace> self.workspaces[i]

            # Detector output is an array so flatten it and assign to correct slice
            if dws.block_size > 1:
                block_i = index * dws.block_size
                block_f = (index + 1) * dws.block_size

                if mask and not dws.ignore_sim_mask:
                    dws.flatiter[block_i : block_f] = np.nan
                else:
                    # TODO (sjr) This relies on get_output returning a ndarray currently,
                    #            it would be good to change this a bit so that a memoryview
                    #            is also accepted (and flattened automatically).
                    dws.flatiter[block_i : block_f] = dws.get_output().flat[:]

            else:
                if mask and not dws.ignore_sim_mask:
                    dws.flatiter[index] = np.nan
                else:
                    dws.flatiter[index] = dws.get_output()

        self._num += 1

    def plot(
        self,
        detectors=None,
        log=False,
        logx=None,
        logy=None,
        degrees=True,
        cmap=None,
        figsize_scale=1,
        tight_layout=True,
        show=True,
        separate=True,
        _test_fig_handles=None,
    ):
        """
        See :mod:`finesse.plotting.plot`.
        """
        plotter = Plotter(self)
        plotter.scale = figsize_scale
        plotter.tight_layout = tight_layout

        return plotter.plot(detectors, log, logx, logy, degrees, cmap, show, separate, _test_fig_handles)

    def get_legacy_data(self, model):
        """Get legacy style data

        This produces a Numpy array and set of headers which is equivelent to parsing a
        Finesse 2 output file.

        See Also
        --------
        :meth:`.write_legacy_data`

        Parameters
        ----------
        model : finesse.model.Model
           The Model used to produce this output file.

        Returns
        -------
        legacy_data : :class:`numpy.ndarray`
            The data array

        column_names : :class:`list`
            List of column names

        plot_type : '2D plot' or '3D plot'
            String indicating if the data should represent 2D or 3D scan.
        """
        import finesse

        # This first block creates the x axes in a Finesse 2 style

        axes = []
        column_names = []
        i = 1

        # Create a list with the same length
        # as the number of axes (i.e. 1 for xaxis, 2 for x2axis etc)
        while True:
            try:
                axes.append(getattr(self, f"x{i}"))
                column_names.append(f"x{i}")
                i += 1
            except IndexError:
                break

        outfile = [[] for ax in axes]
        for idx in np.ndindex(self.shape):
            for ax, val in enumerate(idx):
                outfile[ax].append(axes[ax][val])
        for idx, ax in enumerate(outfile):
            outfile[idx] = np.array(ax)

        if not outfile:
            # noxaxis in finesse2 produces a single 0
            outfile = [np.zeros(1)]


        data = []
        def add_yaxis(el, d):

            def check_real(el):
                #real = (
                #    isinstance(el, finesse.detectors.FieldPixel)
                #    or isinstance(el, finesse.detectors.FieldScanLine)
                #    or isinstance(el, finesse.detectors.FieldCamera)
                #    or isinstance(el, finesse.detectors.Gouy)
                #    or isinstance(el, finesse.detectors.BeamPropertyDetector)
                #    or isinstance(el, finesse.detectors.PowerDetector)
                #)
                #if isinstance(el, finesse.detectors.PowerDetectorDemod1):
                #    real = (el.phase.value is not None)
                #if isinstance(el, finesse.detectors.PowerDetectorDemod2):
                #    real = (el.phase2.value is not None)
                return el.dtype != np.complex128

            real = check_real(el)
            axis = model.yaxis or {"axes": ["abs"]}
            if real:
                any_complex = False
                for el2 in model.detectors:
                    any_complex = any_complex or not check_real(el2)

                if "abs" in axis["axes"]:
                    data.append(np.real(d))
                    column_names.append(str(name)+' abs')
                if "deg" in axis["axes"] and any_complex:
                    data.append(np.angle(d, deg=True))
                    column_names.append(str(name)+' deg')
                if "db" in axis["axes"]:
                    data.append(20 * np.log10(np.abs(d)))
                    column_names.append(str(name)+' db')
                if "re" in axis["axes"]:
                    data.append(np.real(d))
                    column_names.append(str(name)+' re')
                if "im" in axis["axes"]:
                    data.append(np.imag(d))
                    column_names.append(str(name)+' im')
            else:
                if "abs" in axis["axes"]:
                    data.append(np.abs(d))
                    column_names.append(str(name)+' abs')
                if "deg" in axis["axes"]:
                    data.append(np.angle(d, deg=True))
                    column_names.append(str(name)+' deg')
                if "db" in axis["axes"]:
                    data.append(20 * np.log10(np.abs(d)))
                    column_names.append(str(name)+' db')
                if "re" in axis["axes"]:
                    data.append(np.real(d))
                    column_names.append(str(name)+' re')
                if "im" in axis["axes"]:
                    data.append(np.imag(d))
                    column_names.append(str(name)+' im')


        plot_type = '2D plot'
        for name, el in model.elements.items():
            if isinstance(el, finesse.detectors.CCDScanLine) or isinstance(
                el, finesse.detectors.FieldScanLine
            ):
                # The scanline is a little weird, as it's really a beam
                # sweep, so we have to do some rearranging of the data.
                axes = [el.x, *axes]
                outfile = [[] for ax in axes]
                d = np.stack(self[name]).T
                for idx in np.ndindex(d.shape):
                    for ax, val in enumerate(idx):
                        outfile[ax].append(axes[ax][val])
                for idx, ax in enumerate(outfile):
                    outfile[idx] = np.array(ax)
                add_yaxis(el, data, np.ravel(d))
                plot_type = '3D plot'
                break

            if isinstance(el, finesse.detectors.CCDCamera) or isinstance(
                el, finesse.detectors.FieldCamera
            ):
                # Same for the camera
                # Finesse 2 seems to put a 2D beam sweep last, even if the sweeps are the
                # first two axes (see e.g. random/astigmatic_mc02.kat)
                axes = [*axes, el.x, el.y]
                outfile = [[] for ax in axes]
                d = np.stack(self[name])
                for idx in np.ndindex(d.shape):
                    for ax, val in enumerate(idx):
                        outfile[ax].append(axes[ax][val])
                for idx, ax in enumerate(outfile):
                    outfile[idx] = np.array(ax)
                for idx, point in enumerate(data):
                    data[idx] = point * np.ones(np.ravel(d).shape)
                add_yaxis(el, np.ravel(d))
                plot_type = '3D plot'
                break
            if isinstance(el, finesse.detectors.Detector):
                d = np.array(self[name]).flatten()
                add_yaxis(el, d)

        outfile.extend(data)

        return np.array(outfile).T, column_names, plot_type


    def write_legacy_data(
        self,
        model,
        filename="data.out",
        legacy_data=None,
        column_names=None,
        plot_type=None
    ):
        """Write Finesse 2 style ASCII output file

        See Also
        --------
        :meth:`.get_legacy_data`

        Parameters
        ----------
        model : :class:`.Model`
           The model used to produce this output file.

        filename : :class:`str`, optional
            The path of the output file.

        legacy_data : :class:`numpy.ndarray`, optional
            The legacy output data to be written.

        column_names : :class:`list`, optional
            The colomn names which correspond to the legacy data.

        plot_type : "2D plot" or "3D plot", optional
            String indicating if the data should represent 2D or 3D scan.

        Notes
        -----
        If any of legacy_data, column_names or plot_type are None then all three will be
        automatically computed.
        """
        from os import path
        import finesse

        filename = path.abspath(path.expanduser(filename))

        if legacy_data is None or column_names is None or plot_type is None:
            if not (legacy_data is None and column_names is None and plot_type is None):
                LOGGER.warning(
                    "legacy_data, column_names and plot_type are being recomputed."
                )
            legacy_data, column_names, plot_type = self.get_legacy_data(model)

        hdr = 'Finesse3 '+'({})\n'.format(finesse.__version__)
        hdr += str(plot_type) + '\n'
        hdr += ', '.join([_n for _n in column_names]) + '\n'
        hdr += '\n'

        np.savetxt(filename, legacy_data, header=hdr,comments=r'%')



from finesse.simulations.basematrix cimport CarrierSignalMatrixSimulation
from finesse.element cimport ElementWorkspace

import numpy as np
cimport numpy as np

ctypedef object (*fptr_c_output)(DetectorWorkspace)


cdef class OutputInformation:
    cdef:
        readonly unicode name
        object __nodes
        object __dtype
        object __dtype_shape
        object __unit
        object __label
        object __needs_fields
        object __needs_trace


cdef class OutputFuncWrapper:
    cdef fptr_c_output func
    @staticmethod
    cdef OutputFuncWrapper make_from_ptr(fptr_c_output f)


cdef class DetectorWorkspace(ElementWorkspace):
    cdef:
        OutputFuncWrapper fn_c  # C output function
        object fn_py  # Python output function
        readonly CarrierSignalMatrixSimulation sim
        readonly OutputInformation oinfo
        np.flatiter flatiter

        Py_ssize_t block_size # Size of output (typically unity)
        bint ignore_sim_mask # Whether to ignore masking and continue call to get_output

    cpdef get_output(self)


cdef class MaskedDetectorWorkspace(DetectorWorkspace):
    cdef:
        Py_ssize_t* unmasked_mode_indices

    cdef readonly:
        np.ndarray unmasked_indices_arr
        Py_ssize_t num_unmasked_HOMs
        bint has_mask

    cdef int setup_mask(self) except -1
    cdef bint hom_in_modes(self, Py_ssize_t hom_idx)



from finesse.cymath cimport complex_t

from finesse.detectors.workspace cimport (
        DetectorWorkspace,
        MaskedDetectorWorkspace,
)
from finesse.element cimport BaseCValues

cdef class PD1Workspace(MaskedDetectorWorkspace):
    cdef public:
        bint output_real
        int dc_node_id
        int ac_node_id
        int[:,::1] homs
        bint is_f_changing
        bint is_phase_changing
        bint is_audio_mixing
        complex_t Aij
        PD1Values v
    cdef:
        double phase_offset
        Py_ssize_t num_pre_set_beats
        Py_ssize_t* pre_set_beats[2] # 2 x num_pre_set_beats

    cpdef update_beats(self)

cdef class PD1Values(BaseCValues):
    cdef public:
        double f
        double phase

cdef object c_pd1_AC_output(DetectorWorkspace dws)

cdef class PD2Workspace(MaskedDetectorWorkspace):
    cdef public:
        bint output_real
        int dc_node_id
        int ac_node_id
        int[:,::1] homs
        bint is_f1_changing
        bint is_f2_changing
        bint is_phase1_changing
        bint is_phase2_changing
        bint is_audio_mixing
        complex_t z1
        complex_t z2
        PD2Values v

cdef class PD2Values(BaseCValues):
    cdef public:
        double f1
        double phase1
        double f2
        double phase2

cdef object c_pd2_AC_output(DetectorWorkspace dws)



#cython: boundscheck=False, wraparound=False, initializedcheck=False

"""Functions operating on complex numbers of type ``double complex``.

This module provides complex number functions as well as the fundamental
complex data type to be used by developers.

.. note::

    For any developer writing a Cython extension which involves complex
    mathematics, you should ``cimport`` the complex data type, which is used
    consistently across the code-base, via::

        from finesse.cymath cimport complex_t

    or::

        from finesse.cymath.complex cimport complex_t

    Both statements are equivalent. This ``complex_t`` typedef corresponds
    exactly to NumPy's ``np.complex128_t`` --- i.e. ``double complex``.

Most of the standard functions of the C ``"complex.h"`` header are exposed
at a C level via this module. Refer to the `Complex Number Arithmetic C reference
<https://en.cppreference.com/w/c/numeric/complex>`_ for the names, arguments
and further details on these functions. One can ``cimport`` such functions
in the same way as cimporting any other C exposed Cython function. For example::

    from finesse.cymath.complex cimport cexp

will ``cimport`` the `cexp <https://en.cppreference.com/w/c/numeric/complex/cexp>`_
function for use on ``complex_t`` data types in another Cython extension.
"""

cdef extern from "complex.h" nogil:
    double complex cexp(double complex z)
    double complex clog(double complex z)
    double cimag(double complex z)
    double creal(double complex z)

cdef extern from "math.h" nogil:
    double fabs(double arg)
    double fmax(double x, double y)

cdef extern from "constants.h":
    double complex COMPLEX_0 # 0 + 0i
    double complex COMPLEX_1 # 1 + 0i


cpdef bint ceq(complex_t z1, complex_t z2) nogil:
    cdef:
        double maximum
        int re = 0
        int im = 0

    maximum = fmax(
        fmax(fabs(creal(z1)), fabs(creal(z2))),
        fmax(fabs(cimag(z1)), fabs(cimag(z2)))
    )

    if maximum < 1e-13: return True

    # Clip denominator at 1 to avoid amplification of
    # difference for small complex number equality checks
    # NOTE (sjr) Leaving a note here just in case this causes issues in
    #            the future, see Issue #252 for reason for this change
    maximum = fmax(maximum, 1.0)

    if creal(z1) == 0.0 and creal(z2) == 0.0:
        re = 1
    else:
        re = fabs(creal(z1) - creal(z2)) / maximum < 1e-13

    if cimag(z1) == 0.0 and cimag(z2) == 0.0:
        im = 1
    else:
        im = fabs(cimag(z1) - cimag(z2)) / maximum < 1e-13

    return re & im


cpdef complex_t cpow_re(complex_t z, double n) nogil:
    if n == 0.0:
        return COMPLEX_1

    if z == COMPLEX_0:
        return COMPLEX_0

    if n == 1.0:
        return z

    return cexp(n * clog(z))


cpdef void couter(complex_t[::1] x, complex_t[::1] y, complex_t[:, ::1] out) nogil:
    cdef:
        Py_ssize_t i, j
        Py_ssize_t N = x.shape[0]
        Py_ssize_t M = y.shape[0]

    for i in range(N):
        for j in range(M):
            out[i][j] = x[i] * y[j]
